@@ -2,7 +2,7 @@
  *
  *  BlueZ - Bluetooth protocol stack for Linux
  *
- *  Copyright (C) 2004-2007  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2004-2008  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -37,6 +37,10 @@
 
 #include <dbus/dbus.h>
 
+#ifdef NEED_DBUS_WATCH_GET_UNIX_FD
+#define dbus_watch_get_unix_fd dbus_watch_get_fd
+#endif
+
 #ifdef HAVE_DBUS_GLIB
 #include <dbus/dbus-glib-lowlevel.h>
 #endif
@@ -61,6 +65,12 @@ struct watch_info {
 	GIOChannel *io;
 	DBusConnection *conn;
 };
+
+struct server_info {
+	guint watch_id;
+	GIOChannel *io;
+	DBusServer *server;
+};
 #endif
 
 struct disconnect_data {
@@ -74,25 +84,34 @@ struct name_callback {
 };
 
 struct name_data {
+	DBusConnection *connection;
 	char *name;
 	GSList *callbacks;
 };
 
-static struct name_data *name_data_find(const char *name)
+static struct name_data *name_data_find(DBusConnection *connection,
+							const char *name)
 {
 	GSList *current;
 
-	for (current = name_listeners; current != NULL; current = current->next) {
+	for (current = name_listeners;
+			current != NULL; current = current->next) {
 		struct name_data *data = current->data;
-		if (strcmp(name, data->name) == 0)
-			return data;
+
+		if (name == NULL && data->name == NULL) {
+			if (connection == data->connection)
+				return data;
+		} else {
+			if (strcmp(name, data->name) == 0)
+				return data;
+		}
 	}
 
 	return NULL;
 }
 
 static struct name_callback *name_callback_find(GSList *callbacks,
-						name_cb_t func, void *user_data)
+					name_cb_t func, void *user_data)
 {
 	GSList *current;
 
@@ -105,79 +124,78 @@ static struct name_callback *name_callback_find(GSList *callbacks,
 	return NULL;
 }
 
+static void name_data_call_and_free(struct name_data *data)
+{
+	GSList *l;
+
+	for (l = data->callbacks; l != NULL; l = l->next) {
+		struct name_callback *cb = l->data;
+		if (cb->func)
+			cb->func(data->name, cb->user_data);
+		g_free(cb);
+	}
+
+	g_slist_free(data->callbacks);
+	g_free(data->name);
+	g_free(data);
+}
+
 static void name_data_free(struct name_data *data)
 {
 	GSList *l;
 
 	for (l = data->callbacks; l != NULL; l = l->next)
-		free(l->data);
+		g_free(l->data);
 
 	g_slist_free(data->callbacks);
-
-	if (data->name)
-		free(data->name);
-
-	free(data);
+	g_free(data->name);
+	g_free(data);
 }
 
-static int name_data_add(const char *name, name_cb_t func, void *user_data)
+static int name_data_add(DBusConnection *connection,
+			const char *name, name_cb_t func, void *user_data)
 {
 	int first = 1;
 	struct name_data *data = NULL;
 	struct name_callback *cb = NULL;
 
-	cb = malloc(sizeof(struct name_callback));
-	if (!cb)
-		goto failed;
+	cb = g_new(struct name_callback, 1);
 
 	cb->func = func;
 	cb->user_data = user_data;
 
-	data = name_data_find(name);
+	data = name_data_find(connection, name);
 	if (data) {
 		first = 0;
 		goto done;
 	}
 
-	data = malloc(sizeof(struct name_data));
-	if (!data)
-		goto failed;
+	data = g_new0(struct name_data, 1);
 
-	memset(data, 0, sizeof(struct name_data));
-
-	data->name = strdup(name);
-	if (!data->name)
-		goto failed;
+	data->connection = connection;
+	data->name = g_strdup(name);
 
 	name_listeners = g_slist_append(name_listeners, data);
 
 done:
 	data->callbacks = g_slist_append(data->callbacks, cb);
 	return first;
-
-failed:
-	if (data)
-		name_data_free(data);
-
-	if (cb)
-		free(cb);
-
-	return 0;
 }
 
-static void name_data_remove(const char *name, name_cb_t func, void *user_data)
+static void name_data_remove(DBusConnection *connection,
+			const char *name, name_cb_t func, void *user_data)
 {
 	struct name_data *data;
 	struct name_callback *cb = NULL;
 
-	data = name_data_find(name);
+	data = name_data_find(connection, name);
 	if (!data)
 		return;
 
 	cb = name_callback_find(data->callbacks, func, user_data);
 	if (cb) {
 		data->callbacks = g_slist_remove(data->callbacks, cb);
-		free(cb);
+		g_free(cb);
 	}
 
 	if (!data->callbacks) {
@@ -186,6 +204,52 @@ static void name_data_remove(const char *name, name_cb_t func, void *user_data)
 	}
 }
 
+static gboolean add_match(DBusConnection *connection, const char *name)
+{
+	DBusError err;
+	char match_string[128];
+
+	snprintf(match_string, sizeof(match_string),
+			"interface=%s,member=NameOwnerChanged,arg0=%s",
+			DBUS_INTERFACE_DBUS, name);
+
+	dbus_error_init(&err);
+
+	dbus_bus_add_match(connection, match_string, &err);
+
+	if (dbus_error_is_set(&err)) {
+		error("Adding match rule \"%s\" failed: %s", match_string,
+				err.message);
+		dbus_error_free(&err);
+		return FALSE;
+	}
+
+	return TRUE;
+}
+
+static gboolean remove_match(DBusConnection *connection, const char *name)
+{
+	DBusError err;
+	char match_string[128];
+
+	snprintf(match_string, sizeof(match_string),
+			"interface=%s,member=NameOwnerChanged,arg0=%s",
+			DBUS_INTERFACE_DBUS, name);
+
+	dbus_error_init(&err);
+
+	dbus_bus_remove_match(connection, match_string, &err);
+
+	if (dbus_error_is_set(&err)) {
+		error("Removing owner match rule for %s failed: %s",
+				name, err.message);
+		dbus_error_free(&err);
+		return FALSE;
+	}
+
+	return TRUE;
+}
+
 static DBusHandlerResult name_exit_filter(DBusConnection *connection,
 					DBusMessage *message, void *user_data)
 {
@@ -210,7 +274,7 @@ static DBusHandlerResult name_exit_filter(DBusConnection *connection,
 	if (*new != '\0')
 		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
 
-	data = name_data_find(name);
+	data = name_data_find(connection, name);
 	if (!data) {
 		error("Got NameOwnerChanged signal for %s which has no listeners", name);
 		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
@@ -224,61 +288,50 @@ static DBusHandlerResult name_exit_filter(DBusConnection *connection,
 	name_listeners = g_slist_remove(name_listeners, data);
 	name_data_free(data);
 
+	remove_match(connection, name);
+
 	return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
 }
 
 int name_listener_add(DBusConnection *connection, const char *name,
 					name_cb_t func, void *user_data)
 {
-	DBusError err;
-	char match_string[128];
 	int first;
 
 	if (!name_listener_initialized) {
-		if (!dbus_connection_add_filter(connection, name_exit_filter, NULL, NULL)) {
+		if (!dbus_connection_add_filter(connection,
+					name_exit_filter, NULL, NULL)) {
 			error("dbus_connection_add_filter() failed");
 			return -1;
 		}
 		name_listener_initialized = 1;
 	}
 
-	first = name_data_add(name, func, user_data);
+	first = name_data_add(connection, name, func, user_data);
 	/* The filter is already added if this is not the first callback
 	 * registration for the name */
 	if (!first)
 		return 0;
 
-	debug("name_listener_add(%s)", name);
+	if (name) {
+		debug("name_listener_add(%s)", name);
 
-	snprintf(match_string, sizeof(match_string),
-			"interface=%s,member=NameOwnerChanged,arg0=%s",
-			DBUS_INTERFACE_DBUS, name);
-
-	dbus_error_init(&err);
-	dbus_bus_add_match(connection, match_string, &err);
-
-	if (dbus_error_is_set(&err)) {
-		error("Adding match rule \"%s\" failed: %s", match_string,
-				err.message);
-		dbus_error_free(&err);
-		name_data_remove(name, func, user_data);
-		return -1;
+		if (!add_match(connection, name)) {
+			name_data_remove(connection, name, func, user_data);
+			return -1;
+		}
 	}
 
 	return 0;
 }
 
 int name_listener_remove(DBusConnection *connection, const char *name,
-				name_cb_t func, void *user_data)
+					name_cb_t func, void *user_data)
 {
 	struct name_data *data;
 	struct name_callback *cb;
-	DBusError err;
-	char match_string[128];
 
-	debug("name_listener_remove(%s)", name);
-
-	data = name_data_find(name);
+	data = name_data_find(connection, name);
 	if (!data) {
 		error("remove_name_listener: no listener for %s", name);
 		return -1;
@@ -291,27 +344,37 @@ int name_listener_remove(DBusConnection *connection, const char *name,
 	}
 
 	data->callbacks = g_slist_remove(data->callbacks, cb);
-	free(cb);
+	g_free(cb);
 
 	/* Don't remove the filter if other callbacks exist */
 	if (data->callbacks)
 		return 0;
 
-	snprintf(match_string, sizeof(match_string),
-			"interface=%s,member=NameOwnerChanged,arg0=%s",
-			DBUS_INTERFACE_DBUS, name);
+	if (name) {
+		debug("name_listener_remove(%s)", name);
 
-	dbus_error_init(&err);
-	dbus_bus_remove_match(connection, match_string, &err);
+		if (!remove_match(connection, name))
+			return -1;
+	}
 
-	if (dbus_error_is_set(&err)) {
-		error("Removing owner match rule for %s failed: %s",
-							name, err.message);
-		dbus_error_free(&err);
+	name_data_remove(connection, name, func, user_data);
+
+	return 0;
+}
+
+int name_listener_indicate_disconnect(DBusConnection *connection)
+{
+	struct name_data *data;
+
+	data = name_data_find(connection, NULL);
+	if (!data) {
+		error("name_listener_indicate_disconnect: no listener found");
 		return -1;
 	}
 
-	name_data_remove(name, func, user_data);
+	debug("name_listener_indicate_disconnect");
+
+	name_data_call_and_free(data);
 
 	return 0;
 }
@@ -371,17 +434,10 @@ dbus_bool_t dbus_bus_get_unix_process_id(DBusConnection *conn, const char *name,
 static DBusHandlerResult disconnect_filter(DBusConnection *conn,
 						DBusMessage *msg, void *data)
 {
-	const char *iface,*method;
 	struct disconnect_data *dc_data = data;
 
-	if (dbus_message_get_type (msg) != DBUS_MESSAGE_TYPE_SIGNAL)
-		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
-
-	iface = dbus_message_get_interface(msg);
-	method = dbus_message_get_member(msg);
-
-	if ((strcmp(iface, DBUS_INTERFACE_LOCAL) == 0) &&
-			(strcmp(method, "Disconnected") == 0)) {
+	if (dbus_message_is_signal(msg,
+			DBUS_INTERFACE_LOCAL, "Disconnected") == TRUE) {
 		error("Got disconnected from the system message bus");
 		dbus_connection_unref(conn);
 		dc_data->disconnect_cb(dc_data->user_data);
@@ -391,6 +447,75 @@ static DBusHandlerResult disconnect_filter(DBusConnection *conn,
 }
 
 #ifndef HAVE_DBUS_GLIB
+static dbus_int32_t server_slot = -1;
+
+static gboolean server_func(GIOChannel *chan, GIOCondition cond, gpointer data)
+{
+	DBusWatch *watch = data;
+	int flags = 0;
+
+	if (cond & G_IO_IN)  flags |= DBUS_WATCH_READABLE;
+	if (cond & G_IO_OUT) flags |= DBUS_WATCH_WRITABLE;
+	if (cond & G_IO_HUP) flags |= DBUS_WATCH_HANGUP;
+	if (cond & G_IO_ERR) flags |= DBUS_WATCH_ERROR;
+
+	dbus_watch_handle(watch, flags);
+
+	return TRUE;
+}
+
+static dbus_bool_t add_server(DBusWatch *watch, void *data)
+{
+	GIOCondition cond = G_IO_HUP | G_IO_ERR;
+	DBusServer *server = data;
+	struct server_info *info;
+	int fd, flags;
+
+	if (!dbus_watch_get_enabled(watch))
+		return TRUE;
+
+	info = g_new(struct server_info, 1);
+
+	fd = dbus_watch_get_unix_fd(watch);
+	info->io = g_io_channel_unix_new(fd);
+	info->server = dbus_server_ref(server);
+
+	dbus_watch_set_data(watch, info, NULL);
+
+	flags = dbus_watch_get_flags(watch);
+
+	if (flags & DBUS_WATCH_READABLE) cond |= G_IO_IN;
+	if (flags & DBUS_WATCH_WRITABLE) cond |= G_IO_OUT;
+
+	info->watch_id = g_io_add_watch(info->io, cond, server_func, watch);
+
+	return TRUE;
+}
+
+static void remove_server(DBusWatch *watch, void *data)
+{
+	struct server_info *info = dbus_watch_get_data(watch);
+
+	dbus_watch_set_data(watch, NULL, NULL);
+
+	if (info) {
+		g_source_remove(info->watch_id);
+		g_io_channel_unref(info->io);
+		dbus_server_unref(info->server);
+		g_free(info);
+	}
+}
+
+static void server_toggled(DBusWatch *watch, void *data)
+{
+	/* Because we just exit on OOM, enable/disable is
+	 * no different from add/remove */
+	if (dbus_watch_get_enabled(watch))
+		add_server(watch, data);
+	else
+		remove_server(watch, data);
+}
+
 static gboolean message_dispatch_cb(void *data)
 {
 	DBusConnection *connection = data;
@@ -434,11 +559,9 @@ static dbus_bool_t add_watch(DBusWatch *watch, void *data)
 	if (!dbus_watch_get_enabled(watch))
 		return TRUE;
 
-	info = malloc(sizeof(struct watch_info));
-	if (info == NULL)
-		return FALSE;
+	info = g_new(struct watch_info, 1);
 
-	fd = dbus_watch_get_fd(watch);
+	fd = dbus_watch_get_unix_fd(watch);
 	info->io = g_io_channel_unix_new(fd);
 	info->conn = dbus_connection_ref(conn);
 
@@ -464,7 +587,7 @@ static void remove_watch(DBusWatch *watch, void *data)
 		g_source_remove(info->watch_id);
 		g_io_channel_unref(info->io);
 		dbus_connection_unref(info->conn);
-		free(info);
+		g_free(info);
 	}
 }
 
@@ -498,18 +621,17 @@ static void timeout_handler_free(void *data)
 		return;
 
 	g_source_remove(handler->id);
-	free(handler);
+	g_free(handler);
 }
 
 static dbus_bool_t add_timeout(DBusTimeout *timeout, void *data)
 {
 	timeout_handler_t *handler;
 
-	if (!dbus_timeout_get_enabled (timeout))
+	if (!dbus_timeout_get_enabled(timeout))
 		return TRUE;
 
-	handler = malloc(sizeof(timeout_handler_t));
-	memset(handler, 0, sizeof(timeout_handler_t));
+	handler = g_new0(timeout_handler_t, 1);
 
 	handler->timeout = timeout;
 	handler->id = g_timeout_add(dbus_timeout_get_interval(timeout),
@@ -522,7 +644,6 @@ static dbus_bool_t add_timeout(DBusTimeout *timeout, void *data)
 
 static void remove_timeout(DBusTimeout *timeout, void *data)
 {
-
 }
 
 static void timeout_toggled(DBusTimeout *timeout, void *data)
@@ -534,8 +655,7 @@ static void timeout_toggled(DBusTimeout *timeout, void *data)
 }
 
 static void dispatch_status_cb(DBusConnection *conn,
-				DBusDispatchStatus new_status,
-				void *data)
+				DBusDispatchStatus new_status, void *data)
 {
 	if (!dbus_connection_get_is_connected(conn))
 		return;
@@ -545,7 +665,47 @@ static void dispatch_status_cb(DBusConnection *conn,
 }
 #endif
 
-DBusConnection *init_dbus(const char *name, void (*disconnect_cb)(void *), void *user_data)
+void setup_dbus_server_with_main_loop(DBusServer *server)
+{
+#ifdef HAVE_DBUS_GLIB
+	debug("Using D-Bus GLib server setup");
+
+	dbus_server_setup_with_g_main(server, NULL);
+#else
+	dbus_server_allocate_data_slot(&server_slot);
+	if (server_slot < 0)
+		return;
+
+	dbus_server_set_data(server, server_slot, server, NULL);
+
+	dbus_server_set_watch_functions(server, add_server, remove_server,
+						server_toggled, server, NULL);
+
+	dbus_server_set_timeout_functions(server, add_timeout, remove_timeout,
+						timeout_toggled, server, NULL);
+#endif
+}
+
+void setup_dbus_with_main_loop(DBusConnection *conn)
+{
+#ifdef HAVE_DBUS_GLIB
+	debug("Using D-Bus GLib connection setup");
+
+	dbus_connection_setup_with_g_main(conn, NULL);
+#else
+	dbus_connection_set_watch_functions(conn, add_watch, remove_watch,
+						watch_toggled, conn, NULL);
+
+	dbus_connection_set_timeout_functions(conn, add_timeout, remove_timeout,
+						timeout_toggled, conn, NULL);
+
+	dbus_connection_set_dispatch_status_function(conn, dispatch_status_cb,
+								conn, NULL);
+#endif
+}
+
+DBusConnection *init_dbus(const char *name,
+				void (*disconnect_cb)(void *), void *user_data)
 {
 	struct disconnect_data *dc_data;
 	DBusConnection *conn;
@@ -561,18 +721,7 @@ DBusConnection *init_dbus(const char *name, void (*disconnect_cb)(void *), void
 		return NULL;
 	}
 
-#ifdef HAVE_DBUS_GLIB
-	dbus_connection_setup_with_g_main(conn, NULL);
-#else
-	dbus_connection_set_watch_functions(conn, add_watch, remove_watch,
-						watch_toggled, conn, NULL);
-
-	dbus_connection_set_timeout_functions(conn, add_timeout, remove_timeout,
-						timeout_toggled, conn, NULL);
-
-	dbus_connection_set_dispatch_status_function(conn, dispatch_status_cb,
-								conn, NULL);
-#endif
+	setup_dbus_with_main_loop(conn);
 
 	if (name) {
 		dbus_error_init(&err);
@@ -595,12 +744,7 @@ DBusConnection *init_dbus(const char *name, void (*disconnect_cb)(void *), void
 	if (!disconnect_cb)
 		return conn;
 
-	dc_data = malloc(sizeof(struct disconnect_data));
-	if (!dc_data) {
-		error("Allocating disconnect data failed");
-		dbus_connection_unref(conn);
-		return NULL;
-	}
+	dc_data = g_new(struct disconnect_data, 1);
 
 	dc_data->disconnect_cb = disconnect_cb;
 	dc_data->user_data = user_data;
@@ -608,9 +752,9 @@ DBusConnection *init_dbus(const char *name, void (*disconnect_cb)(void *), void
 	dbus_connection_set_exit_on_disconnect(conn, FALSE);
 
 	if (!dbus_connection_add_filter(conn, disconnect_filter,
-				dc_data, free)) {
+				dc_data, g_free)) {
 		error("Can't add D-Bus disconnect filter");
-		free(dc_data);
+		g_free(dc_data);
 		dbus_connection_unref(conn);
 		return NULL;
 	}
@@ -618,6 +762,34 @@ DBusConnection *init_dbus(const char *name, void (*disconnect_cb)(void *), void
 	return conn;
 }
 
+DBusConnection *init_dbus_direct(const char *address)
+{
+	DBusConnection *conn;
+	DBusError err;
+
+	dbus_error_init(&err);
+
+	conn = dbus_connection_open(address, &err);
+
+	if (dbus_error_is_set(&err)) {
+		error("Can't connect to message server: %s", err.message);
+		dbus_error_free(&err);
+		return NULL;
+	}
+
+	setup_dbus_with_main_loop(conn);
+
+	dbus_connection_set_exit_on_disconnect(conn, FALSE);
+
+	return conn;
+}
+
+DBusConnection *dbus_bus_system_setup_with_main_loop(const char *name,
+				void (*disconnect_cb)(void *), void *user_data)
+{
+	return init_dbus(name, disconnect_cb, user_data);
+}
+
 static char simple_xml[] = DBUS_INTROSPECT_1_0_XML_DOCTYPE_DECL_NODE "<node></node>";
 
 DBusHandlerResult simple_introspect(DBusConnection *conn,
@@ -668,3 +840,33 @@ int set_nonblocking(int fd)
 
 	return 0;
 }
+
+void register_external_service(DBusConnection *conn, const char *identifier,
+				const char *name, const char *description)
+{
+	DBusMessage *msg, *reply;
+
+	info("Registering service");
+
+	msg = dbus_message_new_method_call("org.bluez", "/org/bluez",
+				"org.bluez.Database", "RegisterService");
+	if (!msg) {
+		error("Can't create service register method");
+		return;
+	}
+
+	dbus_message_append_args(msg, DBUS_TYPE_STRING, &identifier,
+			DBUS_TYPE_STRING, &name,
+			DBUS_TYPE_STRING, &description, DBUS_TYPE_INVALID);
+
+	reply = dbus_connection_send_with_reply_and_block(conn, msg, -1, NULL);
+	if (!reply) {
+		error("Can't register service");
+		return;
+	}
+
+	dbus_message_unref(msg);
+	dbus_message_unref(reply);
+
+	dbus_connection_flush(conn);
+}
