@@ -2,7 +2,8 @@
  *
  *  BlueZ - Bluetooth protocol stack for Linux
  *
- *  Copyright (C) 2004-2007  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2006-2007  Nokia Corporation
+ *  Copyright (C) 2004-2008  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -21,13 +22,11 @@
  *
  */
 
-#ifndef __BLUEZ_DBUS_DATABASE_H
-#define __BLUEZ_DBUS_DATABASE_H
-
 #define DATABASE_INTERFACE "org.bluez.Database"
 
-DBusHandlerResult handle_database_method(DBusConnection *conn, DBusMessage *msg, void *data);
+dbus_bool_t database_init(DBusConnection *conn, const char *path);
+
+DBusHandlerResult database_message(DBusConnection *conn,
+						DBusMessage *msg, void *data);
 
 void set_sdp_server_enable(void);
-
-#endif /* __BLUEZ_DBUS_DATABSE_H */
