@@ -2,7 +2,8 @@
  *
  *  BlueZ - Bluetooth protocol stack for Linux
  *
- *  Copyright (C) 2004-2007  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2006-2007  Nokia Corporation
+ *  Copyright (C) 2004-2008  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -49,36 +50,32 @@
 #include <dbus/dbus.h>
 
 #include "dbus.h"
+#include "dbus-helper.h"
 #include "hcid.h"
 #include "textfile.h"
+#include "adapter.h"
 #include "dbus-hci.h"
 #include "dbus-common.h"
-#include "dbus-adapter.h"
 #include "dbus-error.h"
+#include "error.h"
 #include "dbus-sdp.h"
 #include "sdp-xml.h"
 
-#define MAX_IDENTIFIER_LEN	29	/* "XX:XX:XX:XX:XX:XX/0xYYYYYYYY\0" */
+#define SESSION_TIMEOUT 2000
 #define DEFAULT_XML_BUF_SIZE	1024
 
-typedef struct {
-	uint16_t dev_id;
-	char *dst;
-	void *search_data;
-	get_record_cb_t *cb;
-	void *data;
-} get_record_data_t;
-
 struct transaction_context {
-	DBusConnection *conn;
-	DBusMessage *rq;
-	sdp_session_t *session;
-
-	/* Used for internal async get remote service record implementation */
-	get_record_data_t *call;
+	DBusConnection	*conn;
+	DBusMessage	*rq;
+	sdp_session_t	*session;
+	GIOChannel	*io;
+	guint		io_id;
+	uuid_t		uuid;
+	GSList		*identifiers;
 };
 
 typedef int connect_cb_t(struct transaction_context *t);
+
 struct pending_connect {
 	DBusConnection *conn;
 	DBusMessage *rq;
@@ -86,9 +83,6 @@ struct pending_connect {
 	char *dst;
 	sdp_session_t *session;
 	connect_cb_t *conn_cb;
-
-	/* Used for internal async get remote service record implementation */
-	get_record_data_t *call;
 };
 
 /* FIXME:  move to a common file */
@@ -98,28 +92,164 @@ typedef struct {
 	char            *info_name;
 } sdp_service_t;
 
+struct cached_session {
+	sdp_session_t *session;
+	guint timeout_id;
+	guint io_id;
+};
+
+static GSList *cached_sessions = NULL;
+
+static gboolean session_timeout(gpointer user_data)
+{
+	struct cached_session *s = user_data;
+
+	debug("sdp session timed out. closing");
+
+	cached_sessions = g_slist_remove(cached_sessions, s);
+
+	g_source_remove(s->io_id);
+	sdp_close(s->session);
+	g_free(s);
+
+	return FALSE;
+}
+
+gboolean idle_callback(GIOChannel *io, GIOCondition cond, gpointer user_data)
+{
+	struct cached_session *s = user_data;
+
+	if (cond & G_IO_NVAL)
+		return FALSE;
+
+	if (cond & (G_IO_ERR | G_IO_HUP))
+		debug("idle_callback: session got disconnected");
+
+	if (cond & G_IO_IN)
+		debug("got unexpected input on idle SDP socket");
+
+	cached_sessions = g_slist_remove(cached_sessions, s);
+
+	g_source_remove(s->timeout_id);
+	sdp_close(s->session);
+	g_free(s);
+
+	return FALSE;
+}
+
+static void cache_sdp_session(sdp_session_t *sess, GIOChannel *io)
+{
+	struct cached_session *s;
+
+	s = g_new0(struct cached_session, 1);
+
+	s->session = sess;
+	s->timeout_id = g_timeout_add(SESSION_TIMEOUT, session_timeout, s);
+	s->io_id = g_io_add_watch(io, G_IO_IN | G_IO_HUP | G_IO_ERR | G_IO_NVAL,
+					idle_callback, s);
+
+	cached_sessions = g_slist_append(cached_sessions, s);
+
+	debug("sdp session added to cache");
+}
+
+static int get_bdaddrs(int sock, bdaddr_t *sba, bdaddr_t *dba)
+{
+	struct sockaddr_l2 a;
+	socklen_t len;
+
+	len = sizeof(a);
+	if (getsockname(sock, (struct sockaddr *) &a, &len) < 0) {
+		error("getsockname: %s (%d)", strerror(errno), errno);
+		return -1;
+	}
+
+	bacpy(sba, &a.l2_bdaddr);
+
+	len = sizeof(a);
+	if (getpeername(sock, (struct sockaddr *) &a, &len) < 0) {
+		error("getpeername: %s (%d)", strerror(errno), errno);
+		return -1;
+	}
+
+	bacpy(dba, &a.l2_bdaddr);
+
+	return 0;
+}
+
+static struct cached_session *get_cached_session(bdaddr_t *src, bdaddr_t *dst)
+{
+	GSList *l;
+
+	for (l = cached_sessions; l != NULL; l = l->next) {
+		struct cached_session *s = l->data;
+		int sock = sdp_get_socket(s->session);
+		bdaddr_t sba, dba;
+
+		if (get_bdaddrs(sock, &sba, &dba) < 0)
+			continue;
+
+		if (bacmp(&sba, src) || bacmp(&dba, dst))
+			continue;
+
+		debug("found matching session, removing from list");
+
+		cached_sessions = g_slist_remove(cached_sessions, s);
+
+		return s;
+	}
+
+	return NULL;
+}
+
+static sdp_session_t *get_sdp_session(bdaddr_t *src, bdaddr_t *dst)
+{
+	struct cached_session *s;
+	sdp_session_t *session;
+
+	s = get_cached_session(src, dst);
+	if (!s) {
+		debug("no matching session found. creating a new one");
+		return sdp_connect(src, dst, SDP_NON_BLOCKING);
+	}
+
+	session = s->session;
+
+	g_source_remove(s->timeout_id);
+	g_source_remove(s->io_id);
+	g_free(s);
+
+	return session;
+}
+
 static void append_and_grow_string(void *data, const char *str)
 {
-	sdp_buf_t *buff = (sdp_buf_t *) data;
+	sdp_buf_t *buff = data;
 	int len;
 
 	len = strlen(str);
 
 	if (!buff->data) {
-		buff->buf_size = DEFAULT_XML_BUF_SIZE;
-		buff->data = realloc(buff->data, buff->buf_size);
+		buff->data = malloc(DEFAULT_XML_BUF_SIZE);
 		if (!buff->data)
 			return;
+		buff->buf_size = DEFAULT_XML_BUF_SIZE;
 	}
 
 	/* Grow string */
 	while (buff->buf_size < (buff->data_size + len + 1)) {
-		/* Grow buffer by a factor of 2 */
-		buff->buf_size = (buff->buf_size << 1);
+		void *tmp;
+		uint32_t new_size;
 
-		buff->data = realloc(buff->data, buff->buf_size);
-		if (!buff->data)
+		/* Grow buffer by a factor of 2 */
+		new_size = (buff->buf_size << 1);
+
+		tmp = realloc(buff->data, new_size);
+		if (!tmp)
 			return;
+
+		buff->data = tmp;
+		buff->buf_size = new_size;
 	}
 
 	/* Include the NULL character */
@@ -142,6 +272,7 @@ sdp_service_t sdp_service[] = {
 	{ "fax",	FAX_SVCLASS_ID,			"Fax"			},
 	{ "spp",	SERIAL_PORT_SVCLASS_ID,		"Serial Port"		},
 	{ "hsp",	HEADSET_SVCLASS_ID,		"Headset"		},
+	{ "hfp",	HANDSFREE_SVCLASS_ID,		"Handsfree"		},
 	{ NULL }
 };
 
@@ -169,18 +300,10 @@ static struct pending_connect *pending_connect_new(DBusConnection *conn, DBusMes
 	if (!dst)
 		return NULL;
 
-	c = malloc(sizeof(*c));
-	if (!c)
-		return NULL;
-	memset(c, 0, sizeof(*c));
+	c = g_new0(struct pending_connect, 1);
 
-	if (dst) {
-		c->dst = strdup(dst);
-		if (!c->dst) {
-			free(c);
-			return NULL;
-		}
-	}
+	if (dst)
+		c->dst = g_strdup(dst);
 
 	c->conn = dbus_connection_ref(conn);
 	c->rq = dbus_message_ref(msg);
@@ -194,8 +317,7 @@ static void pending_connect_free(struct pending_connect *c)
 	if (!c)
 		return;
 
-	if (c->dst)
-		free(c->dst);
+	g_free(c->dst);
 
 	if (c->rq)
 		dbus_message_unref(c->rq);
@@ -203,7 +325,7 @@ static void pending_connect_free(struct pending_connect *c)
 	if (c->conn)
 		dbus_connection_unref(c->conn);
 
-	free(c);
+	g_free(c);
 }
 
 static struct pending_connect *find_pending_connect(const char *dst)
@@ -228,7 +350,7 @@ static const char *get_address_from_message(DBusConnection *conn, DBusMessage *m
 	if (!path)
 		return NULL;
 
-	if (dbus_connection_get_object_path_data(conn, path, (void *) &adapter) == FALSE)
+	if (dbus_connection_get_object_user_data(conn, path, (void *) &adapter) == FALSE)
 		return NULL;
 
 	return adapter->address;
@@ -243,25 +365,21 @@ static int sdp_store_record(const char *src, const char *dst, uint32_t handle, u
 
 	create_file(filename, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
 
-	snprintf(key, sizeof(key), "%s#%08X", dst, handle);
+	snprintf(key, sizeof(key), "%17s#%08X", dst, handle);
 
-	value = malloc(size * 2 + 1);
-	if (!value)
-		return -ENOMEM;
-
-	value[0] = '\0';
+	value = g_malloc0(size * 2 + 1);
 
 	for (i = 0; i < size; i++)
 		sprintf(value + (i * 2), "%02X", buf[i]);
 
 	err = textfile_put(filename, key, value);
 
-	free(value);
+	g_free(value);
 
 	return err;
 }
 
-static void transaction_context_free(void *udata)
+static void transaction_context_free(void *udata, gboolean cache)
 {
 	struct transaction_context *ctxt = udata;
 
@@ -274,47 +392,26 @@ static void transaction_context_free(void *udata)
 	if (ctxt->rq)
 		dbus_message_unref(ctxt->rq);
 
-	if (ctxt->session)
+	if (ctxt->session && !ctxt->io)
 		sdp_close(ctxt->session);
 
-	free(ctxt);
-}
+	if (ctxt->session && ctxt->io) {
+		g_source_remove(ctxt->io_id);
 
-static get_record_data_t *get_record_data_new(uint16_t dev_id, const char *dst,
-					void *search_data,
-					get_record_cb_t *cb, void *data)
-{
-	get_record_data_t *n;
+		if (cache)
+			cache_sdp_session(ctxt->session, ctxt->io);
+		else
+			sdp_close(ctxt->session);
 
-	n = malloc(sizeof(*n));
-	if (!n)
-		return NULL;
-
-	n->dst = strdup(dst);
-	if (!n->dst) {
-		free(n);
-		return NULL;
+		g_io_channel_unref(ctxt->io);
 	}
 
-	n->dev_id = dev_id;
-	n->search_data = search_data;
-	n->cb = cb;
-	n->data = data;
+	if (ctxt->identifiers) {
+		g_slist_foreach(ctxt->identifiers, (GFunc) g_free, NULL);
+		g_slist_free(ctxt->identifiers);
+	}
 
-	return n;
-}
-
-static void get_record_data_free(get_record_data_t *d)
-{
-	free(d->search_data);
-	free(d->dst);
-	free(d);
-}
-
-static inline void get_record_data_call_cb(get_record_data_t *d,
-						sdp_record_t *rec, int err)
-{
-	d->cb(rec, d->data, err);
+	g_free(ctxt);
 }
 
 static gboolean search_process_cb(GIOChannel *chan,
@@ -340,22 +437,18 @@ static gboolean search_process_cb(GIOChannel *chan,
 
 failed:
 	if (err) {
-		if (ctxt->call) {
-			get_record_data_call_cb(ctxt->call, NULL, err);
-			get_record_data_free(ctxt->call);
-		} else
-			error_failed(ctxt->conn, ctxt->rq, err);
-
-		transaction_context_free(ctxt);
+		error_failed_errno(ctxt->conn, ctxt->rq, err);
+		transaction_context_free(ctxt, FALSE);
 	}
 
 	return TRUE;
 }
 
-static void remote_svc_rec_completed_cb(uint8_t type, uint16_t err, uint8_t *rsp, size_t size, void *udata)
+static void remote_svc_rec_completed_cb(uint8_t type, uint16_t err,
+			uint8_t *rsp, size_t size, void *udata)
 {
 	struct transaction_context *ctxt = udata;
-	sdp_record_t *rec = NULL;
+	sdp_record_t *rec;
 	DBusMessage *reply;
 	DBusMessageIter iter, array_iter;
 	const char *src, *dst;
@@ -369,12 +462,12 @@ static void remote_svc_rec_completed_cb(uint8_t type, uint16_t err, uint8_t *rsp
 		int sdp_err = sdp_get_error(ctxt->session);
 		if (sdp_err < 0) {
 			error("search failed: Invalid session!");
-			error_failed(ctxt->conn, ctxt->rq, EINVAL);
+			error_failed_errno(ctxt->conn, ctxt->rq, EINVAL);
 			goto failed;
 		}
 
 		error("search failed: %s (%d)", strerror(sdp_err), sdp_err);
-		error_failed(ctxt->conn, ctxt->rq, sdp_err);
+		error_failed_errno(ctxt->conn, ctxt->rq, sdp_err);
 		goto failed;
 	}
 
@@ -386,7 +479,7 @@ static void remote_svc_rec_completed_cb(uint8_t type, uint16_t err, uint8_t *rsp
 	/* check response PDU ID */
 	if (type != SDP_SVC_ATTR_RSP) {
 		error("SDP error: %s (%d)", strerror(EPROTO), EPROTO);
-		error_failed(ctxt->conn, ctxt->rq, EPROTO);
+		error_failed_errno(ctxt->conn, ctxt->rq, EPROTO);
 		goto failed;
 	}
 
@@ -402,8 +495,8 @@ static void remote_svc_rec_completed_cb(uint8_t type, uint16_t err, uint8_t *rsp
 			DBUS_TYPE_BYTE_AS_STRING, &array_iter);
 
 	rec = sdp_extract_pdu(rsp, &scanned);
-	if (rec == NULL) {
-		error("SVC REC is null");
+	if (rec == NULL || size != scanned) {
+		error("Invalid service record!");
 		goto done;
 	}
 
@@ -419,7 +512,7 @@ done:
 	send_message_and_unref(ctxt->conn, reply);
 
 failed:
-	transaction_context_free(ctxt);
+	transaction_context_free(ctxt, TRUE);
 }
 
 static void remote_svc_rec_completed_xml_cb(uint8_t type, uint16_t err,
@@ -427,7 +520,7 @@ static void remote_svc_rec_completed_xml_cb(uint8_t type, uint16_t err,
 						void *udata)
 {
 	struct transaction_context *ctxt = udata;
-	sdp_record_t *rec = NULL;
+	sdp_record_t *rec;
 	DBusMessage *reply;
 	const char *src, *dst;
 	int scanned;
@@ -441,12 +534,12 @@ static void remote_svc_rec_completed_xml_cb(uint8_t type, uint16_t err,
 		int sdp_err = sdp_get_error(ctxt->session);
 		if (sdp_err < 0) {
 			error("search failed: Invalid session!");
-			error_failed(ctxt->conn, ctxt->rq, EINVAL);
+			error_failed_errno(ctxt->conn, ctxt->rq, EINVAL);
 			goto failed;
 		}
 
 		error("search failed: %s (%d)", strerror(sdp_err), sdp_err);
-		error_failed(ctxt->conn, ctxt->rq, sdp_err);
+		error_failed_errno(ctxt->conn, ctxt->rq, sdp_err);
 		goto failed;
 	}
 
@@ -458,7 +551,7 @@ static void remote_svc_rec_completed_xml_cb(uint8_t type, uint16_t err,
 	/* check response PDU ID */
 	if (type != SDP_SVC_ATTR_RSP) {
 		error("SDP error: %s (%d)", strerror(EPROTO), EPROTO);
-		error_failed(ctxt->conn, ctxt->rq, EPROTO);
+		error_failed_errno(ctxt->conn, ctxt->rq, EPROTO);
 		goto failed;
 	}
 
@@ -469,10 +562,10 @@ static void remote_svc_rec_completed_xml_cb(uint8_t type, uint16_t err,
 	src = get_address_from_message(ctxt->conn, ctxt->rq);
 
 	reply = dbus_message_new_method_return(ctxt->rq);
-	
+
 	rec = sdp_extract_pdu(rsp, &scanned);
-	if (rec == NULL) {
-		error("SVC REC is null");
+	if (rec == NULL || size != scanned) {
+		error("Invalid service record!");
 		goto done;
 	}
 
@@ -495,10 +588,11 @@ done:
 	send_message_and_unref(ctxt->conn, reply);
 
 failed:
-	transaction_context_free(ctxt);
+	transaction_context_free(ctxt, TRUE);
 }
 
-static void remote_svc_handles_completed_cb(uint8_t type, uint16_t err, uint8_t *rsp, size_t size, void *udata)
+static void remote_svc_handles_completed_cb(uint8_t type, uint16_t err,
+			uint8_t *rsp, size_t size, void *udata)
 {
 	struct transaction_context *ctxt = udata;
 	DBusMessage *reply;
@@ -514,12 +608,12 @@ static void remote_svc_handles_completed_cb(uint8_t type, uint16_t err, uint8_t
 		int sdp_err = sdp_get_error(ctxt->session);
 		if (sdp_err < 0) {
 			error("search failed: Invalid session!");
-			error_failed(ctxt->conn, ctxt->rq, EINVAL);
+			error_failed_errno(ctxt->conn, ctxt->rq, EINVAL);
 			goto failed;
 		}
 
 		error("search failed: %s (%d)", strerror(sdp_err), sdp_err);
-		error_failed(ctxt->conn, ctxt->rq, sdp_err);
+		error_failed_errno(ctxt->conn, ctxt->rq, sdp_err);
 		goto failed;
 	}
 
@@ -531,14 +625,14 @@ static void remote_svc_handles_completed_cb(uint8_t type, uint16_t err, uint8_t
 	/* check response PDU ID */
 	if (type != SDP_SVC_SEARCH_RSP) {
 		error("SDP error: %s (%d)", strerror(EPROTO), EPROTO);
-		error_failed(ctxt->conn, ctxt->rq, EPROTO);
+		error_failed_errno(ctxt->conn, ctxt->rq, EPROTO);
 		goto failed;
 	}
 
 	reply = dbus_message_new_method_return(ctxt->rq);
 	dbus_message_iter_init_append(reply, &iter);
 	dbus_message_iter_open_container(&iter, DBUS_TYPE_ARRAY,
-			DBUS_TYPE_UINT32_AS_STRING, &array_iter);
+				DBUS_TYPE_UINT32_AS_STRING, &array_iter);
 
 	pdata = rsp;
 
@@ -559,7 +653,7 @@ static void remote_svc_handles_completed_cb(uint8_t type, uint16_t err, uint8_t
 		pdata += sizeof(uint32_t);
 
 		dbus_message_iter_append_basic(&array_iter,
-				DBUS_TYPE_UINT32, &handle);
+						DBUS_TYPE_UINT32, &handle);
 	} while (--tsrc);
 
 
@@ -568,7 +662,184 @@ done:
 	send_message_and_unref(ctxt->conn, reply);
 
 failed:
-	transaction_context_free(ctxt);
+	transaction_context_free(ctxt, TRUE);
+}
+
+static const char *extract_service_class(sdp_data_t *d)
+{
+	sdp_data_t *seq;
+	uuid_t *uuid;
+	static char uuid_str[37];
+
+	/* Expected sequence of UUID16 */
+	if (d->attrId != SDP_ATTR_SVCLASS_ID_LIST)
+		return NULL;
+
+	if (d->dtd != SDP_SEQ8 && d->dtd != SDP_SEQ16 && d->dtd != SDP_SEQ32)
+		return NULL;
+
+	if (!d->val.dataseq)
+		return NULL;
+
+	seq = d->val.dataseq;
+	if (!SDP_IS_UUID(seq->dtd))
+		return NULL;
+
+	uuid = &seq->val.uuid;
+	if (uuid->type != SDP_UUID16)
+		return NULL;
+
+	sprintf(uuid_str, "0000%04x-0000-1000-8000-00805f9b34fb",
+							uuid->value.uuid16);
+
+	return uuid_str;
+}
+
+static int service_search_attr(struct transaction_context *ctxt, uint16_t uuid)
+{
+	sdp_list_t *attrids, *search;
+	uint32_t range = 0x0000ffff;
+	int ret = 0;
+
+	sdp_uuid16_create(&ctxt->uuid, uuid);
+
+	search = sdp_list_append(0, &ctxt->uuid);
+	attrids = sdp_list_append(NULL, &range);
+
+	/*
+	 * Create/send the search request and set the
+	 * callback to indicate the request completion
+	 */
+	if (sdp_service_search_attr_async(ctxt->session, search,
+				SDP_ATTR_REQ_RANGE, attrids) < 0)
+		ret = -sdp_get_error(ctxt->session);
+
+	sdp_list_free(search, NULL);
+	sdp_list_free(attrids, NULL);
+
+	return ret;
+}
+
+static void remote_svc_identifiers_completed_cb(uint8_t type, uint16_t err,
+			uint8_t *rsp, size_t size, void *udata)
+{
+	struct transaction_context *ctxt = udata;
+	const char *src, *dst, *puuid;
+	const char *devid_uuid = "00001200-0000-1000-8000-00805f9b34fb";
+	char **identifiers;
+	DBusMessage *reply;
+	GSList *l = NULL;
+	int scanned, extracted = 0, len = 0, recsize = 0;
+	uint8_t dtd = 0;
+
+	if (!ctxt)
+		return;
+
+	if (err == 0xffff) {
+		/* Check for protocol error or I/O error */
+		int sdp_err = sdp_get_error(ctxt->session);
+		if (sdp_err < 0) {
+			error("search failed: Invalid session!");
+			error_failed_errno(ctxt->conn, ctxt->rq, EINVAL);
+			goto failed;
+		}
+
+		error("search failed: %s (%d)", strerror(sdp_err), sdp_err);
+		error_failed_errno(ctxt->conn, ctxt->rq, sdp_err);
+		goto failed;
+	}
+
+	if (type == SDP_ERROR_RSP) {
+		error_sdp_failed(ctxt->conn, ctxt->rq, err);
+		goto failed;
+	}
+
+	/* Check response PDU ID */
+	if (type != SDP_SVC_SEARCH_ATTR_RSP) {
+		error("SDP error: %s (%d)", strerror(EPROTO), EPROTO);
+		error_failed_errno(ctxt->conn, ctxt->rq, EPROTO);
+		goto failed;
+	}
+
+	src = get_address_from_message(ctxt->conn, ctxt->rq);
+	dbus_message_get_args(ctxt->rq, NULL,
+			DBUS_TYPE_STRING, &dst,
+			DBUS_TYPE_INVALID);
+
+	scanned = sdp_extract_seqtype(rsp, &dtd, &len);
+	rsp += scanned;
+	for (; extracted < len; rsp += recsize, extracted += recsize) {
+		sdp_record_t *rec;
+		sdp_data_t *d;
+
+		recsize = 0;
+		rec = sdp_extract_pdu(rsp, &recsize);
+		if (!rec)
+			break;
+
+		sdp_store_record(src, dst, rec->handle, rsp, recsize);
+
+		d = sdp_data_get(rec, SDP_ATTR_SVCLASS_ID_LIST);
+		if (!d) {
+			sdp_record_free(rec);
+			continue;
+		}
+
+		puuid = extract_service_class(d);
+		sdp_record_free(rec);
+		if (!puuid)
+			continue;
+
+		/* Ignore repeated identifiers */
+		l = g_slist_find_custom(ctxt->identifiers,
+				puuid, (GCompareFunc) strcmp);
+		if (l)
+			continue;
+
+		ctxt->identifiers = g_slist_append(ctxt->identifiers,
+							g_strdup(puuid));
+	}
+
+	/* If public browse response is empty: search for L2CAP */
+	if (!ctxt->identifiers && ctxt->uuid.value.uuid16 == PUBLIC_BROWSE_GROUP)
+		if (service_search_attr(ctxt, L2CAP_UUID) == 0)
+			return; /* Wait the response */
+
+	/* Request DeviceID if it was not returned previously */
+	l = g_slist_find_custom(ctxt->identifiers,
+			devid_uuid, (GCompareFunc) strcmp);
+	if (!l && ctxt->uuid.value.uuid16 != PNP_INFO_SVCLASS_ID)
+		if (service_search_attr(ctxt, PNP_INFO_SVCLASS_ID) == 0)
+			return; /* Wait the response */
+
+	reply = dbus_message_new_method_return(ctxt->rq);
+
+	identifiers = g_new(char *, g_slist_length(ctxt->identifiers));
+
+	for (l = ctxt->identifiers, len = 0; l; l = l->next, len++)
+		identifiers[len] = l->data;
+
+	dbus_message_append_args(reply,
+			DBUS_TYPE_ARRAY, DBUS_TYPE_STRING,
+			&identifiers, len,
+			DBUS_TYPE_INVALID);
+	send_message_and_unref(ctxt->conn, reply);
+
+	if (len)
+		dbus_connection_emit_signal(ctxt->conn,
+				dbus_message_get_path(ctxt->rq),
+				ADAPTER_INTERFACE,
+				"RemoteIdentifiersUpdated",
+				DBUS_TYPE_STRING, &dst,
+				DBUS_TYPE_ARRAY, DBUS_TYPE_STRING,
+				&identifiers, len,
+				DBUS_TYPE_INVALID);
+
+	if (identifiers)
+		g_free(identifiers);
+
+failed:
+	transaction_context_free(ctxt, TRUE);
 }
 
 static gboolean sdp_client_connect_cb(GIOChannel *chan,
@@ -592,18 +863,11 @@ static gboolean sdp_client_connect_cb(GIOChannel *chan,
 		goto failed;
 	}
 
-	ctxt = malloc(sizeof(*ctxt));
-	if (!ctxt) {
-		err = ENOMEM;
-		goto failed;
-	}
-	memset(ctxt, 0, sizeof(*ctxt));
+	ctxt = g_new0(struct transaction_context, 1);
 
 	ctxt->conn = dbus_connection_ref(c->conn);
 	ctxt->rq = dbus_message_ref(c->rq);
 	ctxt->session = c->session;
-	if (c->call)
-		ctxt->call = c->call;
 
 	/* set the complete transaction callback and send the search request */
 	sdp_err = c->conn_cb(ctxt);
@@ -614,26 +878,21 @@ static gboolean sdp_client_connect_cb(GIOChannel *chan,
 	}
 
 	/* set the callback responsible for update the transaction data */
-	g_io_add_watch(chan, G_IO_IN | G_IO_HUP | G_IO_ERR | G_IO_NVAL,
-			search_process_cb, ctxt);
+	ctxt->io_id = g_io_add_watch(chan,
+				G_IO_IN | G_IO_HUP | G_IO_ERR | G_IO_NVAL,
+				search_process_cb, ctxt);
+	ctxt->io = g_io_channel_ref(chan);
+
 	goto done;
 
 failed:
-	if (c->call)
-		get_record_data_call_cb(c->call, NULL, err);
-	else
-		error_connection_attempt_failed(c->conn, c->rq, err);
-
-	if (c->call)
-		get_record_data_free(c->call);
+	error_connection_attempt_failed(c->conn, c->rq, err);
 
 	if (ctxt)
-		transaction_context_free(ctxt);
+		transaction_context_free(ctxt, FALSE);
 	else
 		sdp_close(c->session);
 
-	g_io_channel_unref(chan);
-
 done:
 	pending_connects = g_slist_remove(pending_connects, c);
 	pending_connect_free(c);
@@ -661,7 +920,7 @@ static struct pending_connect *connect_request(DBusConnection *conn,
 	hci_devba(dev_id, &srcba);
 	str2ba(dst, &dstba);
 
-	c->session = sdp_connect(&srcba, &dstba, SDP_NON_BLOCKING);
+	c->session = get_sdp_session(&srcba, &dstba);
 	if (!c->session) {
 		if (err)
 			*err = errno;
@@ -672,6 +931,7 @@ static struct pending_connect *connect_request(DBusConnection *conn,
 
 	chan = g_io_channel_unix_new(sdp_get_socket(c->session));
 	g_io_add_watch(chan, G_IO_OUT, sdp_client_connect_cb, c);
+	g_io_channel_unref(chan);
 	pending_connects = g_slist_append(pending_connects, c);
 
 	return c;
@@ -679,16 +939,13 @@ static struct pending_connect *connect_request(DBusConnection *conn,
 
 static int remote_svc_rec_conn_cb(struct transaction_context *ctxt)
 {
-	sdp_list_t *attrids = NULL;
+	sdp_list_t *attrids;
 	uint32_t range = 0x0000ffff;
 	const char *dst;
 	uint32_t handle;
-	int err = 0;
 
-	if (sdp_set_notify(ctxt->session, remote_svc_rec_completed_cb, ctxt) < 0) {
-		err = -EINVAL;
-		goto fail;
-	}
+	if (sdp_set_notify(ctxt->session, remote_svc_rec_completed_cb, ctxt) < 0)
+		return -EINVAL;
 
 	dbus_message_get_args(ctxt->rq, NULL,
 			DBUS_TYPE_STRING, &dst,
@@ -696,31 +953,30 @@ static int remote_svc_rec_conn_cb(struct transaction_context *ctxt)
 			DBUS_TYPE_INVALID);
 
 	attrids = sdp_list_append(NULL, &range);
-	/* Create/send the search request and set the callback to indicate the request completion */
-	if (sdp_service_attr_async(ctxt->session, handle, SDP_ATTR_REQ_RANGE, attrids) < 0) {
-		err = -sdp_get_error(ctxt->session);
-		goto fail;
+	/*
+	 * Create/send the search request and set the
+	 * callback to indicate the request completion
+	 */
+	if (sdp_service_attr_async(ctxt->session, handle,
+				SDP_ATTR_REQ_RANGE, attrids) < 0) {
+		sdp_list_free(attrids, NULL);
+		return -sdp_get_error(ctxt->session);
 	}
 
-fail:
-	if (attrids)
-		sdp_list_free(attrids, NULL);
+	sdp_list_free(attrids, NULL);
 
-	return err;
+	return 0;
 }
 
 static int remote_svc_rec_conn_xml_cb(struct transaction_context *ctxt)
 {
-	sdp_list_t *attrids = NULL;
+	sdp_list_t *attrids;
 	uint32_t range = 0x0000ffff;
 	const char *dst;
 	uint32_t handle;
-	int err = 0;
 
-	if (sdp_set_notify(ctxt->session, remote_svc_rec_completed_xml_cb, ctxt) < 0) {
-		err = -EINVAL;
-		goto fail;
-	}
+	if (sdp_set_notify(ctxt->session, remote_svc_rec_completed_xml_cb, ctxt) < 0)
+		return -EINVAL;
 
 	dbus_message_get_args(ctxt->rq, NULL,
 			DBUS_TYPE_STRING, &dst,
@@ -728,32 +984,38 @@ static int remote_svc_rec_conn_xml_cb(struct transaction_context *ctxt)
 			DBUS_TYPE_INVALID);
 
 	attrids = sdp_list_append(NULL, &range);
-	/* Create/send the search request and set the callback to indicate the request completion */
-	if (sdp_service_attr_async(ctxt->session, handle, SDP_ATTR_REQ_RANGE, attrids) < 0) {
-		err = -sdp_get_error(ctxt->session);
-		goto fail;
+	/* 
+	 * Create/send the search request and set the
+	 * callback to indicate the request completion
+	 */
+	if (sdp_service_attr_async(ctxt->session, handle,
+				SDP_ATTR_REQ_RANGE, attrids) < 0) {
+		sdp_list_free(attrids, NULL);
+		return -sdp_get_error(ctxt->session);
 	}
 
-fail:
-	if (attrids)
-		sdp_list_free(attrids, NULL);
+	sdp_list_free(attrids, NULL);
 
-	return err;
+	return 0;
 }
 
-DBusHandlerResult get_remote_svc_rec(DBusConnection *conn, DBusMessage *msg, void *data, sdp_format_t format)
+DBusHandlerResult get_remote_svc_rec(DBusConnection *conn, DBusMessage *msg,
+				void *data, sdp_format_t format)
 {
 	struct adapter *adapter = data;
 	const char *dst;
 	uint32_t handle;
-	int err = 0;
+	int err;
 	connect_cb_t *cb;
 
+	if (!adapter->up)
+		return error_not_ready(conn, msg);
+
 	if (!dbus_message_get_args(msg, NULL,
 			DBUS_TYPE_STRING, &dst,
 			DBUS_TYPE_UINT32, &handle,
 			DBUS_TYPE_INVALID))
-		return error_invalid_arguments(conn, msg);
+		return error_invalid_arguments(conn, msg, NULL);
 
 	if (find_pending_connect(dst))
 		return error_service_search_in_progress(conn, msg);
@@ -765,7 +1027,7 @@ DBusHandlerResult get_remote_svc_rec(DBusConnection *conn, DBusMessage *msg, voi
 	if (!connect_request(conn, msg, adapter->dev_id,
 				dst, cb, &err)) {
 		error("Search request failed: %s (%d)", strerror(err), err);
-		return error_failed(conn, msg, err);
+		return error_failed_errno(conn, msg, err);
 	}
 
 	return DBUS_HANDLER_RESULT_HANDLED;
@@ -775,13 +1037,9 @@ static int remote_svc_handles_conn_cb(struct transaction_context *ctxt)
 {
 	sdp_list_t *search = NULL;
 	const char *dst, *svc;
-	uuid_t uuid;
-	int err = 0;
 
-	if (sdp_set_notify(ctxt->session, remote_svc_handles_completed_cb, ctxt) < 0) {
-		err = -EINVAL;
-		goto fail;
-	}
+	if (sdp_set_notify(ctxt->session, remote_svc_handles_completed_cb, ctxt) < 0)
+		return -EINVAL;
 
 	dbus_message_get_args(ctxt->rq, NULL,
 			DBUS_TYPE_STRING, &dst,
@@ -789,44 +1047,54 @@ static int remote_svc_handles_conn_cb(struct transaction_context *ctxt)
 			DBUS_TYPE_INVALID);
 
 	if (strlen(svc) > 0)
-		str2uuid(&uuid, svc);
+		str2uuid(&ctxt->uuid, svc);
 	else
-		sdp_uuid16_create(&uuid, PUBLIC_BROWSE_GROUP);
+		sdp_uuid16_create(&ctxt->uuid, PUBLIC_BROWSE_GROUP);
 
-	search = sdp_list_append(0, &uuid);
+	search = sdp_list_append(0, &ctxt->uuid);
 
 	/* Create/send the search request and set the callback to indicate the request completion */
 	if (sdp_service_search_async(ctxt->session, search, 64) < 0) {
 		error("send request failed: %s (%d)", strerror(errno), errno);
-		err = -sdp_get_error(ctxt->session);
-		goto fail;
+		sdp_list_free(search, NULL);
+		return -sdp_get_error(ctxt->session);
 	}
 
-fail:
-	if (search)
-		sdp_list_free(search, NULL);
+	sdp_list_free(search, NULL);
 
-	return err;
+	return 0;
+}
+
+static int remote_svc_identifiers_conn_cb(struct transaction_context *ctxt)
+{
+	if (sdp_set_notify(ctxt->session,
+			remote_svc_identifiers_completed_cb, ctxt) < 0)
+		return -EINVAL;
+
+	return service_search_attr(ctxt, PUBLIC_BROWSE_GROUP);
 }
 
 DBusHandlerResult get_remote_svc_handles(DBusConnection *conn, DBusMessage *msg, void *data)
 {
 	struct adapter *adapter = data;
 	const char *dst, *svc;
-	int err = 0;
+	int err;
 	uuid_t uuid;
 
+	if (!adapter->up)
+		return error_not_ready(conn, msg);
+
 	if (!dbus_message_get_args(msg, NULL,
 			DBUS_TYPE_STRING, &dst,
 			DBUS_TYPE_STRING, &svc,
 			DBUS_TYPE_INVALID))
-		return error_invalid_arguments(conn, msg);
+		return error_invalid_arguments(conn, msg, NULL);
 
 	if (strlen(svc) > 0) {
 		/* Check if it is a service name string */
 		if (str2uuid(&uuid, svc) < 0) {
 			error("Invalid service class name");
-			return error_invalid_arguments(conn, msg);
+			return error_invalid_arguments(conn, msg, NULL);
 		}
 	}
 
@@ -836,253 +1104,65 @@ DBusHandlerResult get_remote_svc_handles(DBusConnection *conn, DBusMessage *msg,
 	if (!connect_request(conn, msg, adapter->dev_id,
 				dst, remote_svc_handles_conn_cb, &err)) {
 		error("Search request failed: %s (%d)", strerror(err), err);
-		return error_failed(conn, msg, err);
+		return error_failed_errno(conn, msg, err);
 	}
 
 	return DBUS_HANDLER_RESULT_HANDLED;
 }
 
-/*
- * Internal async get remote service record implementation
- */
-
-static void get_rec_with_handle_comp_cb(uint8_t type, uint16_t err,
-					uint8_t *rsp, size_t size, void *udata)
+DBusHandlerResult get_remote_svc_identifiers(DBusConnection *conn, DBusMessage *msg, void *data)
 {
-	struct transaction_context *ctxt = udata;
-	int scanned, cb_err = 0;
-	sdp_record_t *rec = NULL;
-
-	if (err == 0xffff) {
-		int sdp_err = sdp_get_error(ctxt->session);
-		if (sdp_err < 0) {
-			error("search failed: Invalid session!");
-			cb_err = EINVAL;
-			goto failed;
-		}
-		error("search failed :%s (%d)", strerror(sdp_err), sdp_err);
-		cb_err = sdp_err;
-		goto failed;
-	}
-
-	if (type == SDP_ERROR_RSP || type != SDP_SVC_ATTR_RSP) {
-		error("SDP error: %s(%d)", strerror(EPROTO), EPROTO);
-		cb_err = EPROTO;
-		goto failed;
-	}
-
-	rec = sdp_extract_pdu(rsp, &scanned);
-	if (!rec) {
-		error("Service record is NULL");
-		cb_err = EPROTO;
-		goto failed;
-	}
-
-failed:
-	get_record_data_call_cb(ctxt->call, rec, cb_err);
-
-	if (rec)
-		sdp_record_free(rec);
-
-	get_record_data_free(ctxt->call);
-
-	transaction_context_free(ctxt);
-}
-
-static int get_rec_with_handle_conn_cb(struct transaction_context *ctxt)
-{
-	uint32_t range = 0x0000ffff;
-	sdp_list_t *attrids = NULL;
-	uint32_t handle;
-	int err = 0;
-
-	if (sdp_set_notify(ctxt->session,
-				get_rec_with_handle_comp_cb, ctxt) < 0) {
-		error("Invalid session data!");
-		err = -EINVAL;
-		goto failed;
-	}
-
-	handle = *((uint32_t *)ctxt->call->search_data);
-	attrids = sdp_list_append(NULL, &range);
-
-	if (sdp_service_attr_async(ctxt->session, handle,
-					SDP_ATTR_REQ_RANGE, attrids) < 0) {
-		error("send request failed: %s (%d)", strerror(errno), errno);
-		err = -errno;
-		goto failed;
-	}
-
-failed:
-	if (attrids)
-		sdp_list_free(attrids, NULL);
-
-	return err;
-}
-
-int get_record_with_handle(DBusConnection *conn, DBusMessage *msg,
-			uint16_t dev_id, const char *dst,
-			uint32_t handle, get_record_cb_t *cb, void *data)
-{
-	struct pending_connect *c;
-	get_record_data_t *d;
-	uint32_t *rec_handle;
+	struct adapter *adapter = data;
+	const char *dst;
 	int err;
 
-	if (find_pending_connect(dst)) {
-		error("SDP search in progress!");
-		return -EINPROGRESS;
-	}
+	if (!adapter->up)
+		return error_not_ready(conn, msg);
 
-	rec_handle = malloc(sizeof(uint32_t));
-	if (!rec_handle)
-		return -ENOMEM;
+	if (!dbus_message_get_args(msg, NULL,
+			DBUS_TYPE_STRING, &dst,
+			DBUS_TYPE_INVALID))
+		return error_invalid_arguments(conn, msg, NULL);
 
-	*rec_handle = handle;
+	if (find_pending_connect(dst))
+		return error_service_search_in_progress(conn, msg);
 
-	d = get_record_data_new(dev_id, dst, rec_handle, cb, data);
-	if (!d) {
-		free(rec_handle);
-		return -ENOMEM;
-	}
-
-	if (!(c = connect_request(conn, msg, dev_id, dst,
-				get_rec_with_handle_conn_cb, &err))) {
+	if (!connect_request(conn, msg, adapter->dev_id,
+				dst, remote_svc_identifiers_conn_cb, &err)) {
 		error("Search request failed: %s (%d)", strerror(err), err);
-		get_record_data_free(d);
-		return -err;
+		return error_failed_errno(conn, msg, err);
 	}
 
-	c->call = d;
-
-	return 0;
+	return DBUS_HANDLER_RESULT_HANDLED;
 }
 
-static void get_rec_with_uuid_comp_cb(uint8_t type, uint16_t err,
-					uint8_t *rsp, size_t size, void *udata)
+DBusHandlerResult finish_remote_svc_transact(DBusConnection *conn,
+						DBusMessage *msg, void *data)
 {
-	struct transaction_context *ctxt = udata;
-	get_record_data_t *d = ctxt->call;
-	int csrc, tsrc, cb_err = 0;
-	uint32_t *handle;
-	uint8_t *pdata;
+	struct cached_session *s;
+	const char *address;
+	struct adapter *adapter = data;
+	DBusMessage *reply;
+	bdaddr_t sba, dba;
 
-	if (err == 0xffff) {
-		int sdp_err = sdp_get_error(ctxt->session);
-		if (sdp_err < 0) {
-			error("search failed: Invalid session!");
-			cb_err = EINVAL;
-			goto failed;
-		}
-		error("search failed: %s (%d)", strerror(sdp_err), sdp_err);
-		cb_err = sdp_err;
-		goto failed;
+	if (!dbus_message_get_args(msg, NULL,
+			DBUS_TYPE_STRING, &address,
+			DBUS_TYPE_INVALID))
+		return error_invalid_arguments(conn, msg, NULL);
+
+	reply = dbus_message_new_method_return(msg);
+	if (!reply)
+		return DBUS_HANDLER_RESULT_NEED_MEMORY;
+
+	str2ba(adapter->address, &sba);
+	str2ba(address, &dba);
+
+	while ((s = get_cached_session(&sba, &dba))) {
+		sdp_close(s->session);
+		g_source_remove(s->timeout_id);
+		g_source_remove(s->io_id);
+		g_free(s);
 	}
 
-	if (type == SDP_ERROR_RSP || type != SDP_SVC_SEARCH_RSP) {
-		error("SDP error: %s (%d)", strerror(EPROTO), EPROTO);
-		cb_err = EPROTO;
-		goto failed;
-	}
-
-	pdata = rsp;
-	tsrc = ntohs(bt_get_unaligned((uint16_t *) pdata));
-	if (tsrc <= 0)
-		goto failed;
-	pdata += sizeof(uint16_t);
-
-	csrc = ntohs(bt_get_unaligned((uint16_t *) pdata));
-	if (csrc <= 0)
-		goto failed;
-	pdata += sizeof(uint16_t);
-
-	handle = malloc(sizeof(*handle));
-	if (!handle) {
-		cb_err = ENOMEM;
-		goto failed;
-	}
-	*handle = ntohl(bt_get_unaligned((uint32_t*)pdata));
-
-	free(d->search_data);
-	d->search_data = handle;
-
-	cb_err = get_rec_with_handle_conn_cb(ctxt);
-	if (cb_err)
-		goto failed;
-
-	return;
-
-failed:
-	get_record_data_call_cb(d, NULL, cb_err);
-
-	get_record_data_free(d);
-
-	transaction_context_free(ctxt);
-}
-
-static int get_rec_with_uuid_conn_cb(struct transaction_context *ctxt)
-{
-	get_record_data_t *d = ctxt->call;
-	sdp_list_t *search = NULL;
-	uuid_t *uuid;
-	int err = 0;
-
-	if (sdp_set_notify(ctxt->session,
-			get_rec_with_uuid_comp_cb, ctxt) < 0) {
-		err = -EINVAL;
-		goto failed;
-	}
-
-	uuid = (uuid_t *)d->search_data;
-	search = sdp_list_append(NULL, uuid);
-
-	if (sdp_service_search_async(ctxt->session, search, 1) < 0) {
-		error("send request failed: %s (%d)", strerror(errno), errno);
-		err = -sdp_get_error(ctxt->session);
-		goto failed;
-	}
-
-failed:
-	if (search)
-		sdp_list_free(search, NULL);
-
-	return err;
-}
-
-int get_record_with_uuid(DBusConnection *conn, DBusMessage *msg,
-			uint16_t dev_id, const char *dst,
-			const uuid_t *uuid, get_record_cb_t *cb, void *data)
-{
-	struct pending_connect *c;
-	get_record_data_t *d;
-	int err;
-	uuid_t *sdp_uuid;
-
-	if (find_pending_connect(dst)) {
-		error("SDP search in progress!");
-		return -EINPROGRESS;
-	}
-
-	sdp_uuid = malloc(sizeof(uuid_t));
-	if (!sdp_uuid)
-		return -ENOMEM;
-
-	memcpy(sdp_uuid, uuid, sizeof(uuid_t));
-
-	d = get_record_data_new(dev_id, dst, sdp_uuid, cb, data);
-	if (!d) {
-		free(sdp_uuid);
-		return -ENOMEM;
-	}
-
-	if (!(c = connect_request(conn, msg, dev_id, dst,
-				get_rec_with_uuid_conn_cb, &err))) {
-		error("Search request failed: %s (%d)", strerror(err), err);
-		get_record_data_free(d);
-		return -err;
-	}
-
-	c->call = d;
-
-	return 0;
+	return send_message_and_unref(conn, reply);
 }
