@@ -2,7 +2,8 @@
  *
  *  BlueZ - Bluetooth protocol stack for Linux
  *
- *  Copyright (C) 2004-2007  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2006-2007  Nokia Corporation
+ *  Copyright (C) 2004-2008  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -21,33 +22,21 @@
  *
  */
 
-#ifndef __BLUEZ_DBUS_SDP_H
-#define __BLUEZ_DBUS_SDP_H
-
-#include <stdint.h>
-#include <bluetooth/bluetooth.h>
-#include <bluetooth/sdp.h>
-
 typedef enum {
 	SDP_FORMAT_XML,
 	SDP_FORMAT_BINARY
 } sdp_format_t;
 
-DBusHandlerResult get_remote_svc_handles(DBusConnection *conn, DBusMessage *msg, void *data);
+DBusHandlerResult get_remote_svc_handles(DBusConnection *conn,
+						DBusMessage *msg, void *data);
 
-DBusHandlerResult get_remote_svc_rec(DBusConnection *conn, DBusMessage *msg, void *data, sdp_format_t format);
+DBusHandlerResult get_remote_svc_identifiers(DBusConnection *conn,
+						DBusMessage *msg, void *data);
+
+DBusHandlerResult get_remote_svc_rec(DBusConnection *conn, DBusMessage *msg,
+					void *data, sdp_format_t format);
+
+DBusHandlerResult finish_remote_svc_transact(DBusConnection *conn,
+						DBusMessage *msg, void *data);
 
 uint16_t sdp_str2svclass(const char *str);
-
-typedef void get_record_cb_t(sdp_record_t *rec, void *data, int err);
-
-int get_record_with_uuid(DBusConnection *conn, DBusMessage *msg,
-			uint16_t dev_id, const char *dst,
-			const uuid_t *uuid, get_record_cb_t *cb, void *data);
-
-int get_record_with_handle(DBusConnection *conn, DBusMessage *msg,
-			uint16_t dev_id, const char *dst,
-			uint32_t handle, get_record_cb_t *cb, void *data);
-
-
-#endif /* __BLUEZ_DBUS_SDP_H */
