@@ -4,7 +4,7 @@
  *
  *  Copyright (C) 2000-2001  Qualcomm Incorporated
  *  Copyright (C) 2002-2003  Maxim Krasnyansky <maxk@qualcomm.com>
- *  Copyright (C) 2002-2007  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2002-2008  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -37,6 +37,8 @@
 #include <sys/stat.h>
 #include <sys/ioctl.h>
 #include <sys/socket.h>
+#include <sys/types.h>
+#include <sys/wait.h>
 
 #include <bluetooth/bluetooth.h>
 #include <bluetooth/hci.h>
@@ -44,10 +46,14 @@
 
 #include <glib.h>
 
+#include <dbus/dbus.h>
+
 #include "notify.h"
 
 #include "hcid.h"
 #include "sdpd.h"
+#include "server.h"
+#include "adapter.h"
 #include "dbus-common.h"
 #include "dbus-service.h"
 #include "dbus-database.h"
@@ -57,12 +63,14 @@ struct hcid_opts hcid;
 struct device_opts default_device;
 struct device_opts *parser_device;
 static struct device_list *device_list = NULL;
+static int child_pipe[2];
 
 static inline void init_device_defaults(struct device_opts *device_opts)
 {
 	memset(device_opts, 0, sizeof(*device_opts));
 	device_opts->scan = SCAN_PAGE;
-	device_opts->name = strdup("BlueZ");
+	device_opts->mode = MODE_CONNECTABLE;
+	device_opts->name = g_strdup("BlueZ");
 	device_opts->discovto = HCID_DEFAULT_DISCOVERABLE_TIMEOUT;
 }
 
@@ -70,19 +78,19 @@ struct device_opts *alloc_device_opts(char *ref)
 {
 	struct device_list *device;
 
-	device = malloc(sizeof(struct device_list));
+	device = g_try_new(struct device_list, 1);
 	if (!device) {
 		info("Can't allocate devlist opts buffer: %s (%d)",
 							strerror(errno), errno);
 		exit(1);
 	}
 
-	device->ref = strdup(ref);
+	device->ref = g_strdup(ref);
 	device->next = device_list;
 	device_list = device;
 
 	memcpy(&device->opts, &default_device, sizeof(struct device_opts));
-	device->opts.name = strdup(default_device.name);
+	device->opts.name = g_strdup(default_device.name);
 
 	return &device->opts;
 }
@@ -91,17 +99,13 @@ static void free_device_opts(void)
 {
 	struct device_list *device, *next;
 
-	if (default_device.name) {
-		free(default_device.name);
-		default_device.name = NULL;
-	}
+	g_free(default_device.name);
 
 	for (device = device_list; device; device = next) {
-		free(device->ref);
-		if (device->opts.name)
-			free(device->opts.name);
+		g_free(device->ref);
+		g_free(device->opts.name);
 		next = device->next;
-		free(device);
+		g_free(device);
 	}
 
 	device_list = NULL;
@@ -124,7 +128,7 @@ static struct device_opts *get_device_opts(int hdev)
 	struct hci_dev_info di;
 
 	/* First try to get BD_ADDR based settings ... */
-	if (!hci_devinfo(hdev, &di) < 0) {
+	if (hci_devinfo(hdev, &di) == 0) {
 		char addr[18];
 		ba2str(&di.bdaddr, addr);
 		device_opts = find_device_opts(addr);
@@ -144,7 +148,7 @@ static struct device_opts *get_device_opts(int hdev)
 	return device_opts;
 }
 
-uint8_t get_startup_mode(int hdev)
+static struct device_opts *get_opts(int hdev)
 {
 	struct device_opts *device_opts = NULL;
 	struct hci_dev_info di;
@@ -152,13 +156,13 @@ uint8_t get_startup_mode(int hdev)
 	int sock;
 
 	if (hdev < 0)
-		return SCAN_DISABLED;
+		return NULL;
 
 	sock = hci_open_dev(hdev);
 	if (sock < 0)
 		goto no_address;
 
-	if (!hci_devinfo(hdev, &di) < 0) {
+	if (hci_devinfo(hdev, &di) < 0) {
 		close(sock);
 		goto no_address;
 	}
@@ -178,9 +182,27 @@ no_address:
 	if (!device_opts)
 		device_opts = &default_device;
 
+	return device_opts;
+}
+
+uint8_t get_startup_scan(int hdev)
+{
+	struct device_opts *device_opts = get_opts(hdev);
+	if (!device_opts)
+		return SCAN_DISABLED;
+
 	return device_opts->scan;
 }
 
+uint8_t get_startup_mode(int hdev)
+{
+	struct device_opts *device_opts = get_opts(hdev);
+	if (!device_opts)
+		return MODE_OFF;
+
+	return device_opts->mode;
+}
+
 int get_discoverable_timeout(int hdev)
 {
 	struct device_opts *device_opts = NULL;
@@ -195,7 +217,7 @@ int get_discoverable_timeout(int hdev)
 	if (sock < 0)
 		goto no_address;
 
-	if (!hci_devinfo(hdev, &di) < 0) {
+	if (hci_devinfo(hdev, &di) < 0) {
 		close(sock);
 		goto no_address;
 	}
@@ -221,8 +243,61 @@ no_address:
 	return device_opts->discovto;
 }
 
-/* 
- * Device name expansion 
+void update_service_classes(const bdaddr_t *bdaddr, uint8_t value)
+{
+	struct hci_dev_list_req *dl;
+	struct hci_dev_req *dr;
+	int i, sk;
+
+	sk = socket(AF_BLUETOOTH, SOCK_RAW, BTPROTO_HCI);
+	if (sk < 0)
+		return;
+
+	dl = g_malloc0(HCI_MAX_DEV * sizeof(*dr) + sizeof(*dl));
+
+	dl->dev_num = HCI_MAX_DEV;
+	dr = dl->dev_req;
+
+	if (ioctl(sk, HCIGETDEVLIST, dl) < 0) {
+		close(sk);
+		g_free(dl);
+		return;
+	}
+
+	dr = dl->dev_req;
+
+	for (i = 0; i < dl->dev_num; i++, dr++) {
+		struct hci_dev_info di;
+		uint8_t cls[3];
+		int dd;
+
+		if (hci_devinfo(dr->dev_id, &di) < 0)
+			continue;
+
+		if (hci_test_bit(HCI_RAW, &di.flags))
+			continue;
+
+		if (get_device_class(di.dev_id, cls) < 0)
+			continue;
+
+		dd = hci_open_dev(di.dev_id);
+		if (dd < 0)
+			continue;
+
+		set_service_classes(dd, cls, value);
+
+		hci_close_dev(dd);
+
+		update_adapter(di.dev_id);
+	}
+
+	g_free(dl);
+
+	close(sk);
+}
+
+/*
+ * Device name expansion
  *   %d - device id
  */
 static char *expand_name(char *dst, int size, char *str, int dev_id)
@@ -279,6 +354,32 @@ static char *expand_name(char *dst, int size, char *str, int dev_id)
 	return dst;
 }
 
+static gboolean child_exit(GIOChannel *io, GIOCondition cond, void *user_data)
+{
+	int status, fd = g_io_channel_unix_get_fd(io);
+	pid_t child_pid;
+
+	if (read(fd, &child_pid, sizeof(child_pid)) != sizeof(child_pid)) {
+		error("child_exit: unable to read child pid from pipe");
+		return TRUE;
+	}
+
+	if (waitpid(child_pid, &status, 0) != child_pid)
+		error("waitpid(%d) failed", child_pid);
+	else
+		debug("child %d exited", child_pid);
+
+	return TRUE;
+}
+
+static void at_child_exit(void)
+{
+	pid_t pid = getpid();
+
+	if (write(child_pipe[1], &pid, sizeof(pid)) != sizeof(pid))
+		error("unable to write to child pipe");
+}
+
 static void configure_device(int dev_id)
 {
 	struct device_opts *device_opts;
@@ -301,22 +402,37 @@ static void configure_device(int dev_id)
 
 	/* Set scan mode */
 	if (read_device_mode(&di.bdaddr, mode, sizeof(mode)) == 0) {
-		if (!strcmp(mode, MODE_OFF) && hcid.offmode == HCID_OFFMODE_NOSCAN)
+		if (!strcmp(mode, "off") && hcid.offmode == HCID_OFFMODE_NOSCAN) {
+			device_opts->mode = MODE_OFF;
 			device_opts->scan = SCAN_DISABLED;
-		else if (!strcmp(mode, MODE_CONNECTABLE))
+		} else if (!strcmp(mode, "connectable")) {
+			device_opts->mode = MODE_CONNECTABLE;
 			device_opts->scan = SCAN_PAGE;
-		else if (!strcmp(mode, MODE_DISCOVERABLE)) {
+		} else if (!strcmp(mode, "discoverable")) {
 			/* Set discoverable only if timeout is 0 */
-			if (!get_discoverable_timeout(dev_id))
+			if (!get_discoverable_timeout(dev_id)) {
 				device_opts->scan = SCAN_PAGE | SCAN_INQUIRY;
-			else
+				device_opts->mode = MODE_DISCOVERABLE;
+			} else {
 				device_opts->scan = SCAN_PAGE;
+				device_opts->mode = MODE_CONNECTABLE;
+			}
+		} else if (!strcmp(mode, "limited")) {
+			/* Set discoverable only if timeout is 0 */
+			if (!get_discoverable_timeout(dev_id)) {
+				device_opts->scan = SCAN_PAGE | SCAN_INQUIRY;
+				device_opts->mode = MODE_LIMITED;
+			} else {
+				device_opts->scan = SCAN_PAGE;
+				device_opts->mode = MODE_CONNECTABLE;
+			}
 		}
 	}
 
 	/* Do configuration in the separate process */
 	switch (fork()) {
 		case 0:
+			atexit(at_child_exit);
 			break;
 		case -1:
 			error("Fork failed. Can't init device hci%d: %s (%d)",
@@ -329,7 +445,7 @@ static void configure_device(int dev_id)
 	if (dd < 0) {
 		error("Can't open device hci%d: %s (%d)",
 						dev_id, strerror(errno), errno);
-		return;
+		exit(1);
 	}
 
 	memset(&dr, 0, sizeof(dr));
@@ -365,35 +481,13 @@ static void configure_device(int dev_id)
 	/* Set device name */
 	if ((device_opts->flags & (1 << HCID_SET_NAME)) && device_opts->name) {
 		change_local_name_cp cp;
-		write_ext_inquiry_response_cp ip;
-		char name[249];
-		uint8_t len;
 
-		memset(name, 0, sizeof(name));
-		if (read_local_name(&di.bdaddr, name) < 0) {
-			memset(cp.name, 0, sizeof(cp.name));
-			expand_name((char *) cp.name, sizeof(cp.name),
+		memset(cp.name, 0, sizeof(cp.name));
+		expand_name((char *) cp.name, sizeof(cp.name),
 						device_opts->name, dev_id);
-		} else
-			memcpy(cp.name, name, sizeof(cp.name));
-
-		ip.fec = 0x00;
-		memset(ip.data, 0, sizeof(ip.data));
-		len = strlen((char *) cp.name);
-		if (len > 48) {
-			len = 48;
-			ip.data[1] = 0x08;
-		} else
-			ip.data[1] = 0x09;
-		ip.data[0] = len + 1;
-		memcpy(ip.data + 2, cp.name, len);
 
 		hci_send_cmd(dd, OGF_HOST_CTL, OCF_CHANGE_LOCAL_NAME,
 					CHANGE_LOCAL_NAME_CP_SIZE, &cp);
-
-		if (di.features[6] & LMP_EXT_INQ)
-			hci_send_cmd(dd, OGF_HOST_CTL, OCF_WRITE_EXT_INQUIRY_RESPONSE,
-					WRITE_EXT_INQUIRY_RESPONSE_CP_SIZE, &ip);
 	}
 
 	/* Set device class */
@@ -404,9 +498,14 @@ static void configure_device(int dev_id)
 
 		if (read_local_class(&di.bdaddr, cls) < 0) {
 			class = htobl(device_opts->class);
+			cls[2] = get_service_classes(&di.bdaddr);
 			memcpy(cp.dev_class, &class, 3);
-		} else
+		} else {
+			if (!(device_opts->scan & SCAN_INQUIRY))
+				cls[1] &= 0xdf; /* Clear discoverable bit */
+			cls[2] = get_service_classes(&di.bdaddr);
 			memcpy(cp.dev_class, cls, 3);
+		}
 
 		hci_send_cmd(dd, OGF_HOST_CTL, OCF_WRITE_CLASS_OF_DEV,
 					WRITE_CLASS_OF_DEV_CP_SIZE, &cp);
@@ -441,6 +540,7 @@ static void init_device(int dev_id)
 	/* Do initialization in the separate process */
 	switch (fork()) {
 		case 0:
+			atexit(at_child_exit);
 			break;
 		case -1:
 			error("Fork failed. Can't init device hci%d: %s (%d)",
@@ -460,45 +560,56 @@ static void init_device(int dev_id)
 	if (ioctl(dd, HCIDEVUP, dev_id) < 0 && errno != EALREADY) {
 		error("Can't init device hci%d: %s (%d)",
 					dev_id, strerror(errno), errno);
-		exit(1);
+		goto fail;
 	}
 
 	if (hci_devinfo(dev_id, &di) < 0)
-		exit(1);
+		goto fail;
 
 	if (hci_test_bit(HCI_RAW, &di.flags))
-		exit(0);
+		goto done;
 
 	if (hcid.offmode == HCID_OFFMODE_DEVDOWN) {
 		char mode[16];
 
 		if (read_device_mode(&di.bdaddr, mode, sizeof(mode)) == 0 &&
-						strcmp(mode, MODE_OFF) == 0) {
+						strcmp(mode, "off") == 0) {
 			ioctl(dd, HCIDEVDOWN, dev_id);
-			exit(0);
+			goto done;
 		}
 	}
 
+done:
 	hci_close_dev(dd);
-
 	exit(0);
+
+fail:
+	hci_close_dev(dd);
+	exit(1);
 }
 
 static void device_devreg_setup(int dev_id)
 {
+	struct hci_dev_info di;
+
 	if (hcid.auto_init)
 		init_device(dev_id);
-	hcid_dbus_register_device(dev_id);
+
+	if (hci_devinfo(dev_id, &di) < 0)
+		return;
+
+	if (!hci_test_bit(HCI_RAW, &di.flags))
+		hcid_dbus_register_device(dev_id);
 }
 
 static void device_devup_setup(int dev_id)
 {
-	add_device(dev_id);
+	add_adapter(dev_id);
 	if (hcid.auto_init)
 		configure_device(dev_id);
 	if (hcid.security)
 		start_security_manager(dev_id);
-	start_device(dev_id);
+	start_adapter(dev_id);
 	hcid_dbus_start_device(dev_id);
 }
 
@@ -508,14 +619,13 @@ static void init_all_devices(int ctl)
 	struct hci_dev_req *dr;
 	int i;
 
-	dl = malloc(HCI_MAX_DEV * sizeof(struct hci_dev_req) + sizeof(uint16_t));
+	dl = g_try_malloc0(HCI_MAX_DEV * sizeof(struct hci_dev_req) + sizeof(uint16_t));
 	if (!dl) {
 		info("Can't allocate devlist buffer: %s (%d)",
 							strerror(errno), errno);
 		exit(1);
 	}
 
-	memset(dl, 0, HCI_MAX_DEV * sizeof(struct hci_dev_req) + sizeof(uint16_t));
 	dl->dev_num = HCI_MAX_DEV;
 	dr = dl->dev_req;
 
@@ -534,7 +644,7 @@ static void init_all_devices(int ctl)
 		}
 	}
 
-	free(dl);
+	g_free(dl);
 }
 
 static void init_defaults(void)
@@ -558,7 +668,7 @@ static inline void device_event(GIOChannel *chan, evt_stack_internal *si)
 	case HCI_DEV_UNREG:
 		info("HCI dev %d unregistered", sd->dev_id);
 		hcid_dbus_unregister_device(sd->dev_id);
-		remove_device(sd->dev_id);
+		remove_adapter(sd->dev_id);
 		break;
 
 	case HCI_DEV_UP:
@@ -571,7 +681,7 @@ static inline void device_event(GIOChannel *chan, evt_stack_internal *si)
 		hcid_dbus_stop_device(sd->dev_id);
 		if (hcid.security)
 			stop_security_manager(sd->dev_id);
-		stop_device(sd->dev_id);
+		stop_adapter(sd->dev_id);
 		break;
 	}
 }
@@ -649,7 +759,7 @@ static void usage(void)
 {
 	printf("hcid - HCI daemon ver %s\n", VERSION);
 	printf("Usage: \n");
-	printf("\thcid [-n] [-d] [-s] [-f config file]\n");
+	printf("\thcid [-n] [-d] [-s] [-m mtu] [-f config file]\n");
 }
 
 int main(int argc, char *argv[])
@@ -657,7 +767,8 @@ int main(int argc, char *argv[])
 	struct sockaddr_hci addr;
 	struct hci_filter flt;
 	struct sigaction sa;
-	GIOChannel *ctl_io;
+	GIOChannel *ctl_io, *child_io;
+	uint16_t mtu = 0;
 	int opt, daemonize = 1, debug = 0, sdp = 0, experimental = 0;
 
 	/* Default HCId settings */
@@ -676,7 +787,7 @@ int main(int argc, char *argv[])
 
 	init_defaults();
 
-	while ((opt = getopt(argc, argv, "ndsxf:")) != EOF) {
+	while ((opt = getopt(argc, argv, "ndsm:xf:")) != EOF) {
 		switch (opt) {
 		case 'n':
 			daemonize = 0;
@@ -690,12 +801,16 @@ int main(int argc, char *argv[])
 			sdp = 1;
 			break;
 
+		case 'm':
+			mtu = atoi(optarg);
+			break;
+
 		case 'x':
 			experimental = 1;
 			break;
 
 		case 'f':
-			hcid.config_file = strdup(optarg);
+			hcid.config_file = g_strdup(optarg);
 			break;
 
 		default:
@@ -725,7 +840,6 @@ int main(int argc, char *argv[])
 	sigaction(SIGUSR2, &sa, NULL);
 
 	sa.sa_handler = SIG_IGN;
-	sigaction(SIGCHLD, &sa, NULL);
 	sigaction(SIGPIPE, &sa, NULL);
 
 	if (debug) {
@@ -761,7 +875,19 @@ int main(int argc, char *argv[])
 	if (read_config(hcid.config_file) < 0)
 		error("Config load failed");
 
-	init_devices();
+	if (pipe(child_pipe) < 0) {
+		error("pipe(): %s (%d)", strerror(errno), errno);
+		exit(1);
+	}
+
+	child_io = g_io_channel_unix_new(child_pipe[0]);
+	g_io_channel_set_close_on_unref(child_io, TRUE);
+	g_io_add_watch(child_io,
+			G_IO_IN | G_IO_ERR | G_IO_HUP | G_IO_NVAL,
+			child_exit, NULL);
+	g_io_channel_unref(child_io);
+
+	init_adapters();
 
 	if (hcid_dbus_init() < 0) {
 		error("Unable to get on D-Bus");
@@ -788,11 +914,14 @@ int main(int argc, char *argv[])
 
 	if (sdp) {
 		set_sdp_server_enable();
-		start_sdp_server(0, SDP_SERVER_COMPAT);
+		start_sdp_server(mtu, hcid.deviceid, SDP_SERVER_COMPAT);
+		set_service_classes_callback(update_service_classes);
 	}
 
 	notify_init();
 
+	init_local_server();
+
 	init_services(CONFIGDIR);
 
 	/* Start event processor */
@@ -807,6 +936,8 @@ int main(int argc, char *argv[])
 
 	notify_close();
 
+	shutdown_local_server();
+
 	cleanup_sdp_session();
 
 	g_main_loop_unref(event_loop);
