@@ -45,26 +45,28 @@
      K_SECURITY = 261,
      K_PAIRING = 262,
      K_OFFMODE = 263,
-     K_PTYPE = 264,
-     K_NAME = 265,
-     K_CLASS = 266,
-     K_VOICE = 267,
-     K_PAGETO = 268,
-     K_LM = 269,
-     K_LP = 270,
-     K_ISCAN = 271,
-     K_PSCAN = 272,
-     K_DISCOVTO = 273,
-     K_PASSKEY = 274,
-     K_YES = 275,
-     K_NO = 276,
-     WORD = 277,
-     PATH = 278,
-     STRING = 279,
-     LIST = 280,
-     HCI = 281,
-     BDADDR = 282,
-     NUM = 283
+     K_DEVICEID = 264,
+     K_PTYPE = 265,
+     K_NAME = 266,
+     K_CLASS = 267,
+     K_VOICE = 268,
+     K_PAGETO = 269,
+     K_LM = 270,
+     K_LP = 271,
+     K_ISCAN = 272,
+     K_PSCAN = 273,
+     K_DISCOVTO = 274,
+     K_PASSKEY = 275,
+     K_YES = 276,
+     K_NO = 277,
+     WORD = 278,
+     PATH = 279,
+     STRING = 280,
+     LIST = 281,
+     HCI = 282,
+     BDADDR = 283,
+     ID = 284,
+     NUM = 285
    };
 #endif
 /* Tokens.  */
@@ -74,39 +76,41 @@
 #define K_SECURITY 261
 #define K_PAIRING 262
 #define K_OFFMODE 263
-#define K_PTYPE 264
-#define K_NAME 265
-#define K_CLASS 266
-#define K_VOICE 267
-#define K_PAGETO 268
-#define K_LM 269
-#define K_LP 270
-#define K_ISCAN 271
-#define K_PSCAN 272
-#define K_DISCOVTO 273
-#define K_PASSKEY 274
-#define K_YES 275
-#define K_NO 276
-#define WORD 277
-#define PATH 278
-#define STRING 279
-#define LIST 280
-#define HCI 281
-#define BDADDR 282
-#define NUM 283
+#define K_DEVICEID 264
+#define K_PTYPE 265
+#define K_NAME 266
+#define K_CLASS 267
+#define K_VOICE 268
+#define K_PAGETO 269
+#define K_LM 270
+#define K_LP 271
+#define K_ISCAN 272
+#define K_PSCAN 273
+#define K_DISCOVTO 274
+#define K_PASSKEY 275
+#define K_YES 276
+#define K_NO 277
+#define WORD 278
+#define PATH 279
+#define STRING 280
+#define LIST 281
+#define HCI 282
+#define BDADDR 283
+#define ID 284
+#define NUM 285
 
 
 
 
 #if ! defined YYSTYPE && ! defined YYSTYPE_IS_DECLARED
 typedef union YYSTYPE
-#line 57 "parser.y"
+#line 59 "parser.y"
 {
 	char *str;
 	long  num;
 }
-/* Line 1529 of yacc.c.  */
-#line 110 "parser.h"
+/* Line 1489 of yacc.c.  */
+#line 114 "parser.h"
 	YYSTYPE;
 # define yystype YYSTYPE /* obsolescent; will be withdrawn */
 # define YYSTYPE_IS_DECLARED 1
