@@ -2,7 +2,7 @@
  *
  *  BlueZ - Bluetooth protocol stack for Linux
  *
- *  Copyright (C) 2003-2007  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2003-2008  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -236,6 +236,16 @@ static int request_encryption(bdaddr_t *src, bdaddr_t *dst)
 	return err;
 }
 
+static void enable_sixaxis(int csk)
+{
+	const unsigned char buf[] = {
+		0x53 /*HIDP_TRANS_SET_REPORT | HIDP_DATA_RTYPE_FEATURE*/,
+		0xf4,  0x42, 0x03, 0x00, 0x00 };
+	int err;
+
+	err = write(csk, buf, sizeof(buf));
+}
+
 static int create_device(int ctl, int csk, int isk, uint8_t subclass, int nosdp, int nocheck, int bootonly, int encrypt, int timeout)
 {
 	struct hidp_connadd_req req;
@@ -324,6 +334,9 @@ create:
 		req.flags |= (1 << HIDP_BOOT_PROTOCOL_MODE);
 	}
 
+	if (req.vendor == 0x054c && req.product == 0x0268)
+		enable_sixaxis(csk);
+
 	err = ioctl(ctl, HIDPCONNADD, &req);
 
 error:
