@@ -2,7 +2,7 @@
  *
  *  BlueZ - Bluetooth protocol stack for Linux
  *
- *  Copyright (C) 2003-2007  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2003-2008  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -321,7 +321,7 @@ static struct {
 	{ 2656, "Unified 21c"	},
 	{ 2658, "Unified 21c"	},
 	{ 3057, "Unified 21d"	},
-	{ 3058, "Unified 21d"	},	
+	{ 3058, "Unified 21d"	},
 	{ 3059, "Unified 21d"	},
 	{ 3060, "Unified 21d"	},
 	{ 3062, "Unified 21d"	},
@@ -333,11 +333,48 @@ static struct {
 	{ 3415, "Unified 21f"	},
 	{ 3424, "Unified 21f"	},
 	{ 3454, "Unified 21f"	},
+	{ 3684, "Unified 21f"	},
+	{ 3764, "Unified 21f"	},
+	{ 4276, "Unified 22b"	},
+	{ 4277, "Unified 22b"	},
+	{ 4279, "Unified 22b"	},
+	{ 4281, "Unified 22b"	},
+	{ 4282, "Unified 22b"	},
+	{ 4283, "Unified 22b"	},
+	{ 4284, "Unified 22b"	},
+	{ 4285, "Unified 22b"	},
+	{ 4289, "Unified 22b"	},
+	{ 4290, "Unified 22b"	},
+	{ 4291, "Unified 22b"	},
+	{ 4292, "Unified 22b"	},
+	{ 4293, "Unified 22b"	},
+	{ 4294, "Unified 22b"	},
+	{ 4295, "Unified 22b"	},
+	{ 4363, "Unified 22c"	},
+	{ 4373, "Unified 22c"	},
+	{ 4374, "Unified 22c"	},
+	{ 4532, "Unified 22d"	},
+	{ 4533, "Unified 22d"	},
+	{ 4698, "Unified 23c"	},
+	{ 4839, "Unified 23c"	},
+	{ 4841, "Unified 23c"	},
+	{ 4866, "Unified 23c"	},
+	{ 4867, "Unified 23c"	},
+	{ 4868, "Unified 23c"	},
+	{ 4869, "Unified 23c"	},
+	{ 4870, "Unified 23c"	},
+	{ 4871, "Unified 23c"	},
+	{ 4872, "Unified 23c"	},
+	{ 4874, "Unified 23c"	},
+	{ 4875, "Unified 23c"	},
+	{ 4876, "Unified 23c"	},
+	{ 4877, "Unified 23c"	},
 	{ 2526, "Marcel 1 (2005-09-26)"	},
 	{ 2543, "Marcel 2 (2005-09-28)"	},
 	{ 2622, "Marcel 3 (2005-10-27)"	},
 	{ 3326, "Marcel 4 (2006-06-16)"	},
 	{ 3612, "Marcel 5 (2006-10-24)"	},
+	{ 4509, "Marcel 6 (2007-06-11)"	},
 	{  195, "Sniff 1 (2001-11-27)"	},
 	{  220, "Sniff 2 (2002-01-03)"	},
 	{  269, "Sniff 3 (2002-02-22)"	},
@@ -348,7 +385,7 @@ static struct {
 	{  306, "Sniff 8 (2002-04-12)"	},
 	{  343, "Sniff 9 (2002-05-02)"	},
 	{  346, "Sniff 10 (2002-05-03)"	},
-	{  355, "Sniff 11 (2002-05-16)" },
+	{  355, "Sniff 11 (2002-05-16)"	},
 	{  256, "Sniff 11 (2002-05-16)"	},
 	{  390, "Sniff 12 (2002-06-26)"	},
 	{  450, "Sniff 13 (2002-08-16)"	},
@@ -384,33 +421,34 @@ static struct {
 	{ 1759, "Sniff 40 (2004-11-03)"	},
 	{ 1760, "Sniff 40 (2004-11-03)"	},
 	{ 1761, "Sniff 40 (2004-11-03)"	},
-	{ 2009, "Sniff 41 (2005-04-06)" },
-	{ 2010, "Sniff 41 (2005-04-06)" },
-	{ 2011, "Sniff 41 (2005-04-06)" },
-	{ 2016, "Sniff 42 (2005-04-11)" },
-	{ 2017, "Sniff 42 (2005-04-11)" },
-	{ 2018, "Sniff 42 (2005-04-11)" },
-	{ 2023, "Sniff 43 (2005-04-14)" },
-	{ 2024, "Sniff 43 (2005-04-14)" },
-	{ 2025, "Sniff 43 (2005-04-14)" },
-	{ 2032, "Sniff 44 (2005-04-18)" },
-	{ 2033, "Sniff 44 (2005-04-18)" },
-	{ 2034, "Sniff 44 (2005-04-18)" },
-	{ 2288, "Sniff 45 (2005-07-08)" },
-	{ 2289, "Sniff 45 (2005-07-08)" },
-	{ 2290, "Sniff 45 (2005-07-08)" },
-	{ 2388, "Sniff 46 (2005-08-17)" },
-	{ 2389, "Sniff 46 (2005-08-17)" },
-	{ 2390, "Sniff 46 (2005-08-17)" },
-	{ 2869, "Sniff 47 (2006-02-15)" },
-	{ 2870, "Sniff 47 (2006-02-15)" },
-	{ 2871, "Sniff 47 (2006-02-15)" },
-	{ 3214, "Sniff 48 (2006-05-16)" },
-	{ 3215, "Sniff 48 (2006-05-16)" },
-	{ 3216, "Sniff 48 (2006-05-16)" },
+	{ 2009, "Sniff 41 (2005-04-06)"	},
+	{ 2010, "Sniff 41 (2005-04-06)"	},
+	{ 2011, "Sniff 41 (2005-04-06)"	},
+	{ 2016, "Sniff 42 (2005-04-11)"	},
+	{ 2017, "Sniff 42 (2005-04-11)"	},
+	{ 2018, "Sniff 42 (2005-04-11)"	},
+	{ 2023, "Sniff 43 (2005-04-14)"	},
+	{ 2024, "Sniff 43 (2005-04-14)"	},
+	{ 2025, "Sniff 43 (2005-04-14)"	},
+	{ 2032, "Sniff 44 (2005-04-18)"	},
+	{ 2033, "Sniff 44 (2005-04-18)"	},
+	{ 2034, "Sniff 44 (2005-04-18)"	},
+	{ 2288, "Sniff 45 (2005-07-08)"	},
+	{ 2289, "Sniff 45 (2005-07-08)"	},
+	{ 2290, "Sniff 45 (2005-07-08)"	},
+	{ 2388, "Sniff 46 (2005-08-17)"	},
+	{ 2389, "Sniff 46 (2005-08-17)"	},
+	{ 2390, "Sniff 46 (2005-08-17)"	},
+	{ 2869, "Sniff 47 (2006-02-15)"	},
+	{ 2870, "Sniff 47 (2006-02-15)"	},
+	{ 2871, "Sniff 47 (2006-02-15)"	},
+	{ 3214, "Sniff 48 (2006-05-16)"	},
+	{ 3215, "Sniff 48 (2006-05-16)"	},
+	{ 3216, "Sniff 48 (2006-05-16)"	},
 	{ 3356, "Sniff 49 (2006-07-17)"	},
 	{ 3529, "Sniff 50 (2006-09-21)"	},
 	{ 3546, "Sniff 51 (2006-09-29)"	},
+	{ 3683, "Sniff 52 (2006-11-03)"	},
 	{    0, }
 };
 
