@@ -4,7 +4,7 @@
  *
  *  Copyright (C) 2000-2001  Qualcomm Incorporated
  *  Copyright (C) 2002-2003  Maxim Krasnyansky <maxk@qualcomm.com>
- *  Copyright (C) 2002-2007  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2002-2008  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -63,6 +63,20 @@ static int dev_info(int s, int dev_id, long arg)
 	return 0;
 }
 
+static char *type2str(uint8_t type)
+{
+	switch (type) {
+	case SCO_LINK:
+		return "SCO";
+	case ACL_LINK:
+		return "ACL";
+	case ESCO_LINK:
+		return "eSCO";
+	default:
+		return "Unknown";
+	}
+}
+
 static int conn_list(int s, int dev_id, long arg)
 {
 	struct hci_conn_list_req *cl;
@@ -90,8 +104,7 @@ static int conn_list(int s, int dev_id, long arg)
 		char addr[18];
 		ba2str(&ci->bdaddr, addr);
 		printf("\t%s %s %s handle %d state %d lm %s\n",
-			ci->out ? "<" : ">",
-			ci->type == ACL_LINK ? "ACL" : "SCO",
+			ci->out ? "<" : ">", type2str(ci->type),
 			addr, ci->handle, ci->state,
 			hci_lmtostr(ci->link_mode));
 	}
@@ -471,6 +484,7 @@ static struct option scan_options[] = {
 	{ "numrsp",	1, 0, 'n' },
 	{ "iac",	1, 0, 'i' },
 	{ "flush",	0, 0, 'f' },
+	{ "refresh",	0, 0, 'r' },
 	{ "class",	0, 0, 'C' },
 	{ "info",	0, 0, 'I' },
 	{ "oui",	0, 0, 'O' },
@@ -481,7 +495,7 @@ static struct option scan_options[] = {
 
 static char *scan_help =
 	"Usage:\n"
-	"\tscan [--length=N] [--numrsp=N] [--iac=lap] [--flush] [--class] [--info] [--oui]\n";
+	"\tscan [--length=N] [--numrsp=N] [--iac=lap] [--flush] [--class] [--info] [--oui] [--refresh]\n";
 
 static void cmd_scan(int dev_id, int argc, char **argv)
 {
@@ -494,7 +508,7 @@ static void cmd_scan(int dev_id, int argc, char **argv)
 	struct hci_version version;
 	struct hci_dev_info di;
 	struct hci_conn_info_req *cr;
-	int extcls = 0, extinf = 0, extoui = 0;
+	int refresh = 0, extcls = 0, extinf = 0, extoui = 0;
 	int i, n, l, opt, dd, cc, nc;
 
 	length  = 8;	/* ~10 seconds */
@@ -530,6 +544,10 @@ static void cmd_scan(int dev_id, int argc, char **argv)
 			flags |= IREQ_CACHE_FLUSH;
 			break;
 
+		case 'r':
+			refresh = 1;
+			break;
+
 		case 'C':
 			extcls = 1;
 			break;
@@ -585,12 +603,15 @@ static void cmd_scan(int dev_id, int argc, char **argv)
 		printf("\n");
 
 	for (i = 0; i < num_rsp; i++) {
-		memset(name, 0, sizeof(name));
-		tmp = get_device_name(&di.bdaddr, &(info+i)->bdaddr);
-		if (tmp) {
-			strncpy(name, tmp, 249);
-			free(tmp);
-			nc = 1;
+		if (!refresh) {
+			memset(name, 0, sizeof(name));
+			tmp = get_device_name(&di.bdaddr, &(info+i)->bdaddr);
+			if (tmp) {
+				strncpy(name, tmp, 249);
+				free(tmp);
+				nc = 1;
+			} else
+				nc = 0;
 		} else
 			nc = 0;
 
@@ -905,9 +926,11 @@ static void cmd_info(int dev_id, int argc, char **argv)
 	}
 
 	if ((di.features[7] & LMP_EXT_FEAT) && (features[7] & LMP_EXT_FEAT)) {
-		if (hci_read_remote_ext_features(dd, handle, 0, &max_page, features, 20000) == 0)
+		if (hci_read_remote_ext_features(dd, handle, 0,
+					&max_page, features, 20000) == 0)
 			if (max_page > 0)
-				printf("\tExtended features: %d pages\n", max_page);
+				printf("\tExtended features: %d page%s\n",
+					max_page, max_page > 1 ? "s" : "");
 	}
 
 	if (cc) {
